"""detecting and forcing symmetry"""
import numpy as np

from scipy.spatial import distance_matrix

from AaronTools import addlogger
from AaronTools.utils.prime_numbers import Primes
from AaronTools.utils.utils import (
    rotation_matrix, mirror_matrix, proj, angle_between_vectors, perp_vector
)


class SymmetryElement:
    def __init__(self, order, center):
        self.order = order
        self.operation = np.identity(3)
        self.translation = center

    def perp_dist(self, coords):
        """distance from each coordinate perpendicular to this symmetry element"""
        return np.zeros(len(coords))

    def apply_operation(self, coords):
        """returns coords with the symmetry operation applied"""
        coords = coords - self.translation
        coords = np.matmul(coords, self.operation)
        coords += self.translation

        return coords

    def apply_operation_without_translation(self, coords):
        """
        returns coords with the symmetry operation applied but without
        translating the coordinates to or from self's center
        """
        coords = np.matmul(coords, self.operation)

        return coords

    def error(self, geom=None, tolerance=None, groups=None, coords=None):
        """
        error in this symmetry element for the given geometry
        either geom or coords and groups must be given
        if groups is not given and geom is, atoms will be grouped by element
        """
        if coords is None:
            coords = geom.coords

        full_coords2 = self.apply_operation(coords)
        error = 0

        # compute distances between the initial coords and
        # the coords after applying the symmetry operation
        # but only calculate distances for atoms that might
        # be symmetry-equivalent (i.e. in the same inital
        # group, which typically is based on what the atom's
        # neighbors are)
        if groups is not None:
            group_names = groups
        else:
            group_names = geom.elements
        for group in set(group_names):
            ndx = (group_names == group).nonzero()[0]
            coords1 = np.take(coords, ndx, axis=0)
            coords2 = np.take(full_coords2, ndx, axis=0)
            dist_mat = distance_matrix(coords1, coords2)

            perp_dist = self.perp_dist(coords1)
            # treat values less than 1 as 1 to avoid numerical nonsense
            perp_dist = np.maximum(perp_dist, np.ones(len(ndx)))

            error_mat = dist_mat / perp_dist
            min_d = max(np.min(error_mat, axis=1))
            if min_d > error:
                error = min_d

        return error
    
    def equivalent_positions(self, coords, groups):
        """
        return an array with the indices that are equivalent after
        applying this operation
        
        for example:
        ndx = element.equivalent_positions(geom.coords, groups)
        coords[ndx] should be equal to element.apply_operation(geom.coords)
        """
        eq_ndx = np.zeros(len(coords), dtype=int)
        init_partitions = dict()
        init_ndx = dict()
        
        for i, (coord, group) in enumerate(zip(coords, groups)):
            init_partitions.setdefault(group, [])
            init_partitions[group].append(coord)
            init_ndx.setdefault(group, [])
            init_ndx[group].append(i)

        for group in init_partitions:
            coords = init_partitions[group]
            new_coords = self.apply_operation(coords)
            
            dist = distance_matrix(coords, new_coords)
            closest_ndx = np.argmin(dist, axis=1)
            for i, (atom, ndx) in enumerate(zip(init_partitions[group], closest_ndx)):
                j = init_ndx[group][i]
                k = init_ndx[group][ndx]
                eq_ndx[j] = k
        
        return eq_ndx
    
    @property
    def trace(self):
        """trace of this symmetry element's matrix"""
        return np.trace(self.operation)


class Identity(SymmetryElement):
    def __init__(self):
        self.translation = np.zeros(3)
        self.operation = np.eye(3)

    def __repr__(self):
        return "E"

    def __lt__(self, other):
        return False


class ProperRotation(SymmetryElement):
    """proper rotation"""
    def __init__(self, center, axis, n, exp=1):
        self.order = n
        self.operation = rotation_matrix(
            2 * np.pi * exp / n,
            axis,
            renormalize=False,
        )
        self.translation = center
        self.axis = axis
        self.n = n
        self.exp = exp

    def __repr__(self):
        if self.exp > 1:
            return "C%i^%i (%5.2f %5.2f %5.2f)" % (
                self.n,
                self.exp,
                *self.axis,
            )
        return "C%i (%5.2f %5.2f %5.2f)" % (
            self.n,
            *self.axis,
        )

    def __lt__(self, other):
        if isinstance(other, Identity) or isinstance(other, InversionCenter):
            return False
        if isinstance(other, ProperRotation):
            if self.n == other.n:
                return self.exp > other.exp
            return self.n < other.n
        return False

    def perp_dist(self, coords):
        v = coords - self.translation
        n = np.dot(v, self.axis)
        p = np.outer(n, self.axis)
        return np.linalg.norm(v - p, axis=1)


class MirrorPlane(SymmetryElement):
    """mirror plane"""
    def __init__(self, center, axis, label=None):
        self.order = 2
        self.translation = center
        self.axis = axis
        self.operation = mirror_matrix(axis)
        self.label = label

    def __repr__(self):
        if self.label:
            return "sigma_%s (%5.2f %5.2f %5.2f)" % (self.label, *self.axis)
        return "sigma (%5.2f %5.2f %5.2f)" % tuple(self.axis)

    def __lt__(self, other):
        if not isinstance(other, MirrorPlane):
            return True
        if self.label and other.label:
            return self.label < other.label
        return True

    def perp_dist(self, coords):
        v = coords - self.translation
        return np.dot(v, self.axis[:, None]).flatten()


class InversionCenter(SymmetryElement):
    """inversion center"""
    def __init__(self, center):
        self.order = 2
        self.operation = -np.identity(3)
        self.translation = center

    def __lt__(self, other):
        if isinstance(other, Identity):
            return True
        return False

    def __repr__(self):
        return "i (%.2f %.2f %.2f)" % (
            *self.translation,
        )

    def perp_dist(self, coords):
        v = coords - self.translation
        return np.linalg.norm(v, axis=1)


class ImproperRotation(SymmetryElement):
    """improper rotation"""
    def __init__(self, center, axis, n, exp=1):
        self.order = n
        self.operation = np.matmul(
            rotation_matrix(
                2 * np.pi * exp / n,
                axis,
                renormalize=False,
            ),
            mirror_matrix(axis)
        )
        self.axis = axis
        self.translation = center
        self.n = n
        self.exp = exp

    def __repr__(self):
        if self.exp > 1:
            return "S%i^%i (%5.2f %5.2f %5.2f)" % (
                self.n,
                self.exp,
                *self.axis,
            )
        return "S%i (%5.2f %5.2f %5.2f)" % (
            self.n,
            *self.axis,
        )

    def __lt__(self, other):
        if (
                isinstance(other, Identity) or
                isinstance(other, ProperRotation) or
                isinstance(other, InversionCenter)
        ):
            return True
        if isinstance(other, ImproperRotation):
            if self.n == other.n:
                return self.exp > other.exp
            return self.n < other.n
        return False

    def perp_dist(self, coords):
        v = coords - self.translation
        n = np.dot(v, self.axis)
        p = np.outer(n, self.axis)
        ax_dist = np.linalg.norm(v - p, axis=1)

        sig_dist = np.dot(v, self.axis[:, None]).flatten()

        return np.minimum(ax_dist, sig_dist)


@addlogger
class PointGroup:

    LOG = None

    def __init__(
            self,
            geom,
            tolerance=0.1,
            max_rotation=6,
            rotation_tolerance=0.01,
            groups=None,
            center=None
    ):
        self.geom = geom
        self.center = center
        if self.center is None:
            self.center = geom.COM()
        self.elements = self.get_symmetry_elements(
            geom,
            tolerance=tolerance,
            max_rotation=max_rotation,
            groups=groups,
            rotation_tolerance=rotation_tolerance,
        )
        self.name = self.determine_point_group(
            rotation_tolerance=rotation_tolerance
        )

    def get_symmetry_elements(
            self,
            geom,
            tolerance=0.1,
            max_rotation=6,
            rotation_tolerance=0.01,
            groups=None,
    ):
        """
        determine what symmetry elements are valid for geom
        geom - Geometry()
        tolerance - maximum error for an element to be valid
        max_rotation - maximum n for Cn (Sn can be 2x this)
        rotation_tolerance - tolerance in radians for angle between
        axes to be for them to be considered parallel/antiparallel/orthogonal

        returns list(SymmetryElement)
        """
        CITATION = "doi:10.1002/jcc.22995"
        self.LOG.citation(CITATION)

        # atoms are grouped based on what they are bonded to
        # if there's not many atoms, don't bother splitting them up
        # based on ranks
        if groups is not None:
            atom_ids = np.array(groups)
            self.initial_groups = atom_ids
        else:
            atom_ids = np.array(
                geom.canonical_rank(
                    update=False,
                    break_ties=False,
                    invariant=False,
                )
            )
            self.initial_groups = atom_ids

        coords = geom.coords

        moments, axes = geom.get_principle_axes()
        axes = axes.T

        valid = [Identity()]

        degeneracy = np.ones(3, dtype=int)
        for i, m1 in enumerate(moments):
            for j, m2 in enumerate(moments):
                if i == j:
                    continue
                if np.isclose(m1, m2, rtol=tolerance, atol=tolerance):
                    degeneracy[i] += 1

        com = self.center

        inver = InversionCenter(com)
        error = inver.error(geom, tolerance, groups=atom_ids)
        if error <= tolerance:
            valid.append(inver)

        if any(np.isclose(m, 0) for m in moments):
            return valid

        ortho_to = []
        for vec, degen in zip(axes, degeneracy):
            if any(d > 1 for d in degeneracy) and degen == 1:
                ortho_to.append(vec)
            elif all(d == 1 for d in degeneracy):
                ortho_to.append(vec)

        # find vectors from COM to each atom
        # these might be proper rotation axes
        atom_axes = geom.coords - com

        # find vectors normal to each pair of atoms
        # these might be normal to a miror plane
        atom_pair_norms = []
        for i, v in enumerate(atom_axes):
            dv = atom_axes - v
            c2 = np.linalg.norm(dv, axis=1) ** 2
            angles = np.arccos(-0.5 * (c2 - 2))
            mask2 = angles > rotation_tolerance
            mask3 = angles < np.pi - rotation_tolerance
            mask = np.logical_and(mask2, mask3)
            
            pair_n = np.cross(v, atom_axes[mask])
            norms = np.linalg.norm(pair_n, axis=1)
            pair_n = np.take(pair_n, np.nonzero(norms), axis=0)[0]
            norms = np.take(norms, np.nonzero(norms), axis=0)
            pair_n /= norms.T
            atom_pair_norms.extend(pair_n.tolist())
        
        atom_pair_norms = np.array(atom_pair_norms)

        # find vectors to the midpoints between each
        # pair of like atoms
        # these might be proper rotations
        atom_pairs = []
        for atom_id in set(atom_ids):
            ndx = (atom_ids == atom_id).nonzero()[0]
            subset_axes = np.take(atom_axes, ndx, axis=0)
            for i, v in enumerate(subset_axes):
                mask = np.ones(len(subset_axes), dtype=bool)
                mask[i] = False
                pair_v = subset_axes[mask] + v
                norms = np.linalg.norm(pair_v, axis=1)
                pair_v = np.take(pair_v, np.nonzero(norms), axis=0)[0]
                norms = np.take(norms, np.nonzero(norms), axis=0)
                pair_v /= norms.T
                atom_pairs.extend(pair_v.tolist())

        atom_pairs = np.array(atom_pairs)
        
        norms = np.linalg.norm(atom_axes, axis=1)
        # don't want axis for an atom that is at the COM (0-vector)
        atom_axes = np.take(atom_axes, np.nonzero(norms), axis=0)[0]

        # normalize
        norms = np.take(norms, np.nonzero(norms))
        atom_axes /= norms.T

        # s = ""
        # for v in atom_axes:
        #     s += ".arrow   %f %f %f   " % tuple(com)
        #     end = com + 2 * v
        #     s += "%f %f %f\n" % tuple(end)
        # with open("test2.bild", "w") as f:
        #     f.write(s)

        # remove parallel/antiparallel axes for single atoms
        # print(atom_axes)
        mask = np.ones(len(atom_axes), dtype=bool)
        for i, v in enumerate(atom_axes):
            if not mask[i]:
                continue
            dv = atom_axes - v
            c2 = np.linalg.norm(dv, axis=1) ** 2
            angles = np.arccos(-0.5 * (c2 - 2))
            # print(", ".join(["%.2f" % a for a in angles]))
            mask2 = angles > rotation_tolerance
            mask3 = angles < np.pi - rotation_tolerance
            mask[:i] *= np.logical_and(mask2, mask3)[:i]
            # print(mask)

        atom_axes = atom_axes[mask]

        # s = ""
        # for v in atom_axes:
        #     s += ".arrow   %f %f %f   " % tuple(com)
        #     end = com + 2 * v
        #     s += "%f %f %f\n" % tuple(end)
        # with open("test2.bild", "w") as f:
        #     f.write(s)

        # remove parallel/antiparallel axes for pairs of atoms
        mask = np.ones(len(atom_pairs), dtype=bool)
        for i, v in enumerate(atom_pairs):
            if not mask[i]:
                continue
            dv = np.delete(atom_pairs, i, axis=0) - v
            c2 = np.linalg.norm(dv, axis=1) ** 2
            angles = np.arccos(-0.5 * (c2 - 2))

            mask2 = angles > rotation_tolerance
            mask3 = angles < np.pi - rotation_tolerance
            mask4 = np.logical_and(mask2, mask3)
            mask[:i] *= mask4[:i]
            mask[i + 1:] *= mask4[i:]

        atom_pairs = atom_pairs[mask]

        # remove parallel/antiparallel norms for pairs of atoms
        mask = np.ones(len(atom_pair_norms), dtype=bool)
        for i, v in enumerate(atom_pair_norms):
            if not mask[i]:
                continue
            dv = np.delete(atom_pair_norms, i, axis=0) - v
            c2 = np.linalg.norm(dv, axis=1) ** 2
            angles = np.arccos(-0.5 * (c2 - 2))
            mask2 = angles > rotation_tolerance
            mask3 = angles < np.pi - rotation_tolerance
            mask4 = np.logical_and(mask2, mask3)
            mask[:i] *= mask4[:i]
            mask[i + 1:] *= mask4[i:]

        atom_pair_norms = atom_pair_norms[mask]

        # s = ""
        # for v in atom_pair_norms:
        #     s += ".arrow   %f %f %f   " % tuple(com)
        #     end = com + 2 * v
        #     s += "%f %f %f\n" % tuple(end)
        # with open("test2.bild", "w") as f:
        #     f.write(s)

        if len(atom_pairs):
            # remove axes for pairs of atoms that are parallel/antiparallel
            # to axes for single atoms
            mask = np.ones(len(atom_pairs), dtype=bool)
            for i, v in enumerate(atom_axes):
                dv = atom_pairs - v
                c2 = np.linalg.norm(dv, axis=1) ** 2
                angles = np.arccos(-0.5 * (c2 - 2))
                mask2 = angles > rotation_tolerance
                mask3 = angles < np.pi - rotation_tolerance
                mask *= np.logical_and(mask2, mask3)

            atom_pairs = atom_pairs[mask]

        if len(atom_pair_norms):
            # remove norms for pairs of atoms that are parallel/antiparallel
            # to axes for single atoms
            mask = np.ones(len(atom_pair_norms), dtype=bool)
            for i, v in enumerate(atom_axes):
                dv = atom_pair_norms - v
                c2 = np.linalg.norm(dv, axis=1) ** 2
                angles = np.arccos(-0.5 * (c2 - 2))
                mask2 = angles > rotation_tolerance
                mask3 = angles < np.pi - rotation_tolerance
                mask *= np.logical_and(mask2, mask3)

            atom_pair_norms = atom_pair_norms[mask]

        # s = ""
        # for v in atom_pair_norms:
        #     s += ".arrow   %f %f %f   " % tuple(com)
        #     end = com + 2 * v
        #     s += "%f %f %f\n" % tuple(end)
        # with open("test2.bild", "w") as f:
        #     f.write(s)

        # remove axes for single atoms that are parallel/antiparallel
        # to moment of inertia axes
        mask = np.ones(len(atom_axes), dtype=bool)
        for i, v in enumerate(axes):
            dv = atom_axes - v
            c2 = np.linalg.norm(dv, axis=1) ** 2
            angles = np.arccos(-0.5 * (c2 - 2))
            mask2 = angles > rotation_tolerance
            mask3 = angles < np.pi - rotation_tolerance
            mask *= np.logical_and(mask2, mask3)
        
        atom_axes = atom_axes[mask]

        # remove axes for pairs of atoms that are parallel/antiparallel
        # to moment of inertia axes
        if len(atom_pairs):
            mask = np.ones(len(atom_pairs), dtype=bool)
            for i, v in enumerate(axes):
                dv = atom_pairs - v
                c2 = np.linalg.norm(dv, axis=1) ** 2
                angles = np.arccos(-0.5 * (c2 - 2))
                mask2 = angles > rotation_tolerance
                mask3 = angles < np.pi - rotation_tolerance
                mask *= np.logical_and(mask2, mask3)

            atom_pairs = atom_pairs[mask]

        # remove norms for pairs of atoms that are parallel/antiparallel
        # to moment of inertia axes
        if len(atom_pair_norms):
            mask = np.ones(len(atom_pair_norms), dtype=bool)
            for i, v in enumerate(axes):
                dv = atom_pair_norms - v
                c2 = np.linalg.norm(dv, axis=1) ** 2
                angles = np.arccos(-0.5 * (c2 - 2))
                mask2 = angles > rotation_tolerance
                mask3 = angles < np.pi - rotation_tolerance
                mask *= np.logical_and(mask2, mask3)

            atom_pair_norms = atom_pair_norms[mask]

        # s = ""
        # for v in atom_pair_norms:
        #     s += ".arrow   %f %f %f   " % tuple(com)
        #     end = com + 2 * v
        #     s += "%f %f %f\n" % tuple(end)
        # with open("test2.bild", "w") as f:
        #     f.write(s)

        # remove axes that are not orthogonal to moments of inertia axes
        if ortho_to:
            mask = np.ones(len(atom_axes), dtype=bool)
            pair_mask = np.ones(len(atom_pairs), dtype=bool)
            pair_mask_norms = np.ones(len(atom_pair_norms), dtype=bool)
            for v in ortho_to:
                dv = atom_axes - v
                c2 = np.linalg.norm(dv, axis=1) ** 2
                angles = np.arccos(-0.5 * (c2 - 2))
                mask1 = abs(angles - np.pi / 2) < rotation_tolerance
                mask *= mask1

        
                if len(atom_pairs):
                    dv = atom_pairs - v
                    c2 = np.linalg.norm(dv, axis=1) ** 2
                    angles = np.arccos(-0.5 * (c2 - 2))
                    pair_mask = abs(angles - np.pi / 2) < rotation_tolerance
                    atom_pairs = atom_pairs[pair_mask]
        
                if len(atom_pair_norms):
                    dv = atom_pair_norms - v
                    c2 = np.linalg.norm(dv, axis=1) ** 2
                    angles = np.arccos(-0.5 * (c2 - 2))
                    pair_mask_norms = abs(angles - np.pi / 2) < rotation_tolerance
                    atom_pair_norms = atom_pair_norms[pair_mask_norms]
        
            atom_axes = atom_axes[mask]
 
        for v in axes:
            mask = np.ones(len(atom_axes), dtype=bool)
            pair_mask = np.ones(len(atom_pairs), dtype=bool)
            pair_mask_norms = np.ones(len(atom_pair_norms), dtype=bool)
                
            dv = atom_axes - v
            c2 = np.linalg.norm(dv, axis=1) ** 2
            angles = np.arccos(-0.5 * (c2 - 2))
            mask1 = angles > rotation_tolerance
            mask2 = angles < np.pi - rotation_tolerance
            mask *= np.logical_and(mask1, mask2)
            if len(atom_pairs):
                dv = atom_pairs - v
                c2 = np.linalg.norm(dv, axis=1) ** 2
                angles = np.arccos(-0.5 * (c2 - 2))
                pair_mask1  = angles > rotation_tolerance
                pair_mask2 = angles < np.pi - rotation_tolerance
                pair_mask *= np.logical_and(pair_mask1, pair_mask2)
                atom_pairs = atom_pairs[pair_mask]

            if len(atom_pair_norms):
                dv = atom_pair_norms - v
                c2 = np.linalg.norm(dv, axis=1) ** 2
                angles = np.arccos(-0.5 * (c2 - 2))
                atom_pair_norms1  = angles > rotation_tolerance
                atom_pair_norms2 = angles < np.pi - rotation_tolerance
                pair_mask_norms *= np.logical_and(atom_pair_norms1, atom_pair_norms2)
                atom_pair_norms = atom_pair_norms[pair_mask_norms]

        if len(atom_pairs) and len(atom_axes):
            for v in atom_pairs:
                mask = np.ones(len(atom_axes), dtype=bool)
                dv = atom_axes - v
                c2 = np.linalg.norm(dv, axis=1) ** 2
                angles = np.arccos(-0.5 * (c2 - 2))
                mask1 = angles > rotation_tolerance
                mask2 = angles < np.pi - rotation_tolerance
                mask *= np.logical_and(mask1, mask2)
                atom_axes = atom_axes[mask]

        # s = ""
        # for v in ortho_to:
        #     s += ".arrow   %f %f %f   " % tuple(com)
        #     end = com + 2 * v
        #     s += "%f %f %f\n" % tuple(end)
        # with open("test2.bild", "w") as f:
        #     f.write(s)

        checked_axes = 0
        # find proper rotations along the axes we've found:
        # * moments of inertia axes
        # * COM -> atom vectors
        # * COM -> midpoint of atom paris
        # also grab axes for checking mirror planes
        check_axes = []
        primes = dict()
        args = tuple([arg for arg in [axes, atom_axes, atom_pairs] if len(arg)])
        principal_axis = None
        for ax in np.concatenate(args):
            max_n = None
            found_n = []
            for n in range(2, max_rotation + 1):
                if n not in primes:
                    primes[n] = Primes.primes_below(n // 2)
                # print(n, primes[n])
                skip = False
                for prime in primes[n]:
                    if n % prime == 0 and prime not in found_n:
                        # print("skipping", n)
                        skip = True
                        break
                # if max_n and max_n % n != 0:
                #     # the highest order proper rotation axis must be
                #     # divisible by all other coincident axes
                #     continue
                # look for C5^2 stuff
                # for exp in range(1, 1 + n // 2):
                for exp in range(1, 2):
                    if exp > 1 and n % exp == 0:
                        # skip things like C4^2 b/c that's just C2
                        continue
                    # see if the error associated with the element is reasonable
                    rot = ProperRotation(com, ax, n, exp)
                    error = rot.error(tolerance, groups=atom_ids, coords=coords)
                    checked_axes += 1
                    if error <= tolerance:
                        # print(geom.atoms[i])
                        # s = ".arrow %f %f %f    " % tuple(com)
                        # end = com + 2 * ax
                        # s += "%f %f %f\n" % tuple(end)
                        # with open("test.bild", "a") as f:
                        #     f.write(s)
                        valid.append(rot)
                        if principal_axis is None or rot.n > principal_axis[0].n:
                            principal_axis = [rot]
                        elif principal_axis is not None and rot.n == principal_axis[0].n:
                            principal_axis.append(rot)
                        found_n.append(n)
                        if n > 2:
                            # for Cn n != 2, add an element that is the same
                            # except the axis of rotation is antiparallel
                            rot2 = ProperRotation(com, -ax, n, exp)
                            valid.append(rot2)
                        if not max_n:
                            max_n = n
                            check_axes.append(ax)
                    elif exp == 1:
                        # can't have Cn^y if you don't have Cn
                        break

        if degeneracy[0] == 3:
            # spherical top molecules need more checks related to C2 axes
            c2_axes = list(
                filter(
                    lambda ele: isinstance(ele, ProperRotation) and ele.n == 2,
                    valid,
                )
            )

            # TODO: replace with array operations like before
            for i, c2_1 in enumerate(c2_axes):
                for c2_2 in c2_axes[:i]:
                    test_axes = []
                    if len(c2_axes) == 3:
                        # T groups - check midpoint
                        for c2_3 in c2_axes[i:]:
                            axis = c2_1.axis + c2_2.axis + c2_3.axis
                            test_axes.append(axis)
                            axis = c2_1.axis + c2_2.axis - c2_3.axis
                            test_axes.append(axis)
                            axis = c2_1.axis - c2_2.axis + c2_3.axis
                            test_axes.append(axis)
                            axis = c2_1.axis - c2_2.axis - c2_3.axis
                            test_axes.append(axis)
                    else:
                        # O, I groups - check cross product
                        test_axes.append(np.cross(c2_1.axis, c2_2.axis))

                    for axis in test_axes:
                        norm = np.linalg.norm(axis)
                        if norm < 1e-5:
                            continue
                        axis /= norm
                        dup = False
                        for element in valid:
                            if isinstance(element, ProperRotation):
                                if 1 - abs(np.dot(element.axis, axis)) < rotation_tolerance:
                                    dup = True
                                    break
                        if dup:
                            continue
                        max_n = None
                        for n in range(max_rotation, 1, -1):
                            if max_n and max_n % n != 0:
                                continue
                            # for exp in range(1, 1 + n // 2):
                            for exp in range(1, 2):
                                if exp > 1 and n % exp == 0:
                                    continue
                                rot = ProperRotation(com, axis, n, exp)
                                checked_axes += 1
                                error = rot.error(tolerance, groups=atom_ids, coords=coords)
                                if error <= tolerance:
                                    if principal_axis is None or rot.n > principal_axis[0].n:
                                        principal_axis = [rot]
                                    elif principal_axis is not None and rot.n == principal_axis[0].n:
                                        principal_axis.append(rot)
                                    valid.append(rot)
                                    if not max_n:
                                        max_n = n
                                        check_axes.append(ax)
                                    if n > 2:
                                        rot2 = ProperRotation(com, -axis, n, exp)
                                        valid.append(rot2)
                                elif exp == 1:
                                    break

        # improper rotations
        # coincident with proper rotations and can be 1x or 2x
        # the order of the proper rotation
        for element in valid:
            if not isinstance(element, ProperRotation):
                continue

            if element.exp != 1:
                continue

            for x in [1, 2]:
                if x * element.n == 2:
                    # S2 is inversion - we already checked i
                    continue

                # for exp in range(1, 1 + (x * element.n) // 2):
                for exp in range(1, 2):
                    if exp > 1 and (x * element.n) % exp == 0:
                        continue

                    for element2 in valid:
                        if isinstance(element2, ImproperRotation):
                            angle = angle_between_vectors(element2.axis, element.axis)
                            if (
                                    element2.exp == exp and
                                    (
                                        angle < rotation_tolerance or
                                        angle > (np.pi - rotation_tolerance)
                                    ) and
                                    element2.n == x * element.n
                            ):
                                break
                    else:

                        imp_rot = ImproperRotation(
                            element.translation,
                            element.axis,
                            x * element.n,
                            exp,
                        )

                        error = imp_rot.error(tolerance, groups=atom_ids, coords=coords)

                        if error <= tolerance:
                            valid.append(imp_rot)
                            rot2 = ImproperRotation(
                                element.translation,
                                -element.axis,
                                x * element.n,
                                exp
                            )
                            valid.append(rot2)

                        elif exp == 1:
                            break

        c2_axes = list(
            filter(
                lambda ele: isinstance(ele, ProperRotation) and ele.n == 2 and ele.exp == 1,
                valid,
            )
        )
        c2_vectors = np.array([c2.axis for c2 in c2_axes])

        sigma_norms = []
        if bool(principal_axis) and len(c2_vectors) and principal_axis[0].n != 2:
            for ax in principal_axis:
                perp = np.cross(ax.axis, c2_vectors)
                norms = np.linalg.norm(perp, axis=1)
                mask = np.nonzero(norms)
                perp = perp[mask]
                norms = norms[mask]
                perp /= norms[:, None]
                sigma_norms.extend(perp)
            
            sigma_norms = np.array(sigma_norms)
            mask = np.ones(len(sigma_norms), dtype=bool)
            for i, v in enumerate(sigma_norms):
                if not mask[i]:
                    continue
                dv = np.delete(sigma_norms, i, axis=0) - v
                c2 = np.linalg.norm(dv, axis=1) ** 2
                angles = np.arccos(-0.5 * (c2 - 2))
                mask2 = angles > rotation_tolerance
                mask3 = angles < np.pi - rotation_tolerance
                mask4 = np.logical_and(mask2, mask3)
                mask[:i] *= mask4[:i]
                mask[i + 1:] *= mask4[i:]
            
            sigma_norms = sigma_norms[mask]

        c2_vectors = np.append(c2_vectors, [-c2.axis for c2 in c2_axes], axis=0)

        # mirror axes
        # for I, O - only check c2 axes
        if (
                degeneracy[0] != 3 or
                not c2_axes or
                (degeneracy[0] == 3 and len(c2_axes) == 3)
        ):
            if len(atom_pair_norms):
                mask = np.ones(len(atom_pair_norms), dtype=bool)
                for i, v in enumerate(axes):
                    dv = atom_pair_norms - v
                    c2 = np.linalg.norm(dv, axis=1) ** 2
                    angles = np.arccos(-0.5 * (c2 - 2))
                    mask2 = angles > rotation_tolerance
                    mask3 = angles < np.pi - rotation_tolerance
                    mask *= np.logical_and(mask2, mask3)

                atom_pair_norms = atom_pair_norms[mask]

                mask = np.ones(len(atom_pair_norms), dtype=bool)
                for i, v in enumerate(check_axes):
                    dv = atom_pair_norms - v
                    c2 = np.linalg.norm(dv, axis=1) ** 2
                    angles = np.arccos(-0.5 * (c2 - 2))
                    mask2 = angles > rotation_tolerance
                    mask3 = angles < np.pi - rotation_tolerance
                    mask *= np.logical_and(mask2, mask3)

                atom_pair_norms = atom_pair_norms[mask]
            
            if check_axes:
                check_axes = np.array(check_axes)
                mask = np.ones(len(check_axes), dtype=bool)
                for i, v in enumerate(axes):
                    dv = check_axes - v
                    c2 = np.linalg.norm(dv, axis=1) ** 2
                    angles = np.arccos(-0.5 * (c2 - 2))
                    mask2 = angles > rotation_tolerance
                    mask3 = angles < np.pi - rotation_tolerance
                    mask *= np.logical_and(mask2, mask3)

                check_axes = check_axes[mask]

                mask = np.ones(len(check_axes), dtype=bool)
                for i, v in enumerate(atom_axes):
                    dv = check_axes - v
                    c2 = np.linalg.norm(dv, axis=1) ** 2
                    angles = np.arccos(-0.5 * (c2 - 2))
                    mask2 = angles > rotation_tolerance
                    mask3 = angles < np.pi - rotation_tolerance
                    mask *= np.logical_and(mask2, mask3)

                check_axes = check_axes[mask]

                mask = np.ones(len(check_axes), dtype=bool)
                for i, v in enumerate(atom_pair_norms):
                    dv = check_axes - v
                    c2 = np.linalg.norm(dv, axis=1) ** 2
                    angles = np.arccos(-0.5 * (c2 - 2))
                    mask2 = angles > rotation_tolerance
                    mask3 = angles < np.pi - rotation_tolerance
                    mask *= np.logical_and(mask2, mask3)

                check_axes = check_axes[mask]
            
            if len(sigma_norms):
                sigma_norms = np.array(sigma_norms)
                mask = np.ones(len(sigma_norms), dtype=bool)
                for i, v in enumerate(axes):
                    dv = sigma_norms - v
                    c2 = np.linalg.norm(dv, axis=1) ** 2
                    angles = np.arccos(-0.5 * (c2 - 2))
                    angles = np.nan_to_num(angles)
                    mask2 = angles > rotation_tolerance
                    mask3 = angles < np.pi - rotation_tolerance
                    mask *= np.logical_and(mask2, mask3)

                sigma_norms = sigma_norms[mask]

                mask = np.ones(len(sigma_norms), dtype=bool)
                for i, v in enumerate(atom_axes):
                    dv = sigma_norms - v
                    c2 = np.linalg.norm(dv, axis=1) ** 2
                    angles = np.arccos(-0.5 * (c2 - 2))
                    mask2 = angles > rotation_tolerance
                    mask3 = angles < np.pi - rotation_tolerance
                    mask *= np.logical_and(mask2, mask3)

                sigma_norms = sigma_norms[mask]

                mask = np.ones(len(sigma_norms), dtype=bool)
                for i, v in enumerate(atom_pair_norms):
                    dv = sigma_norms - v
                    c2 = np.linalg.norm(dv, axis=1) ** 2
                    angles = np.arccos(-0.5 * (c2 - 2))
                    mask2 = angles > rotation_tolerance
                    mask3 = angles < np.pi - rotation_tolerance
                    mask *= np.logical_and(mask2, mask3)

                sigma_norms = sigma_norms[mask]

                mask = np.ones(len(sigma_norms), dtype=bool)
                for i, v in enumerate(check_axes):
                    dv = sigma_norms - v
                    c2 = np.linalg.norm(dv, axis=1) ** 2
                    angles = np.arccos(-0.5 * (c2 - 2))
                    mask2 = angles > rotation_tolerance
                    mask3 = angles < np.pi - rotation_tolerance
                    mask *= np.logical_and(mask2, mask3)

                sigma_norms = sigma_norms[mask]

            # print("axes")
            # for ax in axes:
            #     print(ax)
            #     
            # print("atom_axes")
            # for ax in atom_axes:
            #     print(ax)
            #     
            # print("atom_pair_norms")
            # for ax in atom_pair_norms:
            #     print(ax)
            #     
            # print("check_axes")
            # for ax in check_axes:
            #     print(ax)
            #     
            # print("sigma_norms")
            # for ax in sigma_norms:
            #     print(ax)
                
            args = tuple(
                [arg for arg in [
                    axes, atom_axes, atom_pair_norms, check_axes, sigma_norms
                ] if len(arg)]
            )

            for ax in np.concatenate(args):
                mirror = MirrorPlane(com, ax)
                error = mirror.error(tolerance, groups=atom_ids, coords=coords)
                if error <= tolerance:
                    valid.append(mirror)
        else:
            for i, ax in enumerate(c2_axes):
                mirror = MirrorPlane(com, ax.axis)
                error = mirror.error(tolerance, groups=atom_ids, coords=coords)
                if error <= tolerance:
                    valid.append(mirror)

        min_atoms = None
        if principal_axis:
            for ele in valid:
                if not isinstance(ele, MirrorPlane):
                    continue

                for ax in principal_axis:
                    c2 = np.linalg.norm(ax.axis - ele.axis) ** 2
                    angle = np.arccos(-0.5 * (c2 - 2))
                    if angle < rotation_tolerance or angle > np.pi - rotation_tolerance:
                        # mirror plane normal is parallel to principal_axis
                        ele.label = "h"
                        break
                    elif abs(angle - np.pi / 2) < rotation_tolerance:
                        # mirror plane normal is perpendicular to principal_axis
                        ele.label = "v"
                
                if ele.label == "v":
                    # determine number of atoms in sigma_v in case we need to
                    # differentiate from sigma_d later
                    perp_dist = ele.perp_dist(geom.coords)
                    atoms_contained = sum([1 if abs(d) < tolerance else 0 for d in perp_dist])
                    if min_atoms is None or atoms_contained < min_atoms:
                        min_atoms = atoms_contained

        
        if c2_axes:
            # check for sigma_d
            # should be a sigma_v that bisects two C2 axes
            # if a molecule has different sigma_v planes that
            # bisect C2 axes and contain different numbers of
            # atoms, it is convention to label just the sigma_v
            # planes that contain the fewest atoms as sigma_d
            for ele in valid:
                if not isinstance(ele, MirrorPlane):
                    continue
                if principal_axis and ele.label != "v":
                    continue
                perp_dist = ele.perp_dist(geom.coords)
                atoms_contained = sum([1 if abs(d) < tolerance else 0 for d in perp_dist])
                if min_atoms is not None and atoms_contained != min_atoms:
                    continue
                perp_v = perp_vector(ele.axis)
                c2 = np.linalg.norm(c2_vectors - perp_v, axis=1) ** 2
                angles = np.arccos(-0.5 * (c2 - 2))
                for i, angle1 in enumerate(angles):
                    for j, angle2 in enumerate(angles[:i]):
                        if abs(np.dot(c2_vectors[i], c2_vectors[j])) == 1:
                            continue
                        if abs(angle1 - angle2) < rotation_tolerance:
                            ele.label = "d"
                            break
                    
                    if ele.label == "d":
                        break
            
        # s = ""
        # colors = ["black", "red", "blue", "green", "purple", "yellow", "cyan"]
        # for element in valid:
        #     # if isinstance(element, ProperRotation):
        #     #     s += ".color %s\n" % colors[element.n - 2]
        #     #     s += ".note C%i\n" % element.n
        #     #     s += ".arrow %f %f %f    " % tuple(com)
        #     #     end = com + 2 * np.sqrt(element.n) * element.axis
        #     #     s += "%f %f %f\n" % tuple(end)
        #     if isinstance(element, MirrorPlane):
        #         s += ".arrow %f %f %f    " % tuple(com)
        #         end = com + 2 * element.axis
        #         s += "%f %f %f\n" % tuple(end)
        #
        # with open("test.bild", "w") as f:
        #     f.write(s)

        return valid

    def determine_point_group(self, rotation_tolerance=0.01):
        """
        determines point group of self by looing at self.elements
        rotation_tolerance - tolerance in radians for axes to be
        considered parallel/antiparallel/orthogonal

        returns str for point group name
        """
        moments, axes = self.geom.get_principle_axes()
        linear = False
        if any(np.isclose(m, 0) for m in moments):
            linear = True

        if linear:
            if any(isinstance(ele, InversionCenter) for ele in self.elements):
                return "D_inf_h"
            else:
                return "C_inf_v"

        Cn = dict()
        has_inversion = False
        has_mirror = False
        has_sig_h = False
        has_sig_d = False
        has_sig_v = False
        for ele in self.elements:
            if isinstance(ele, ProperRotation):
                if ele.n not in Cn and ele.exp == 1:
                    Cn[ele.n] = []
                Cn[ele.n].append(ele)
            if isinstance(ele, InversionCenter):
                has_inversion = True
            if isinstance(ele, MirrorPlane):
                has_mirror = True
                if ele.label == "d":
                    has_sig_d = True
                elif ele.label == "h":
                    has_sig_h = True
                elif ele.label == "v":
                    has_sig_v = True

        if Cn:
            max_n = max(Cn.keys())

            if max_n == 5 and len(Cn[5]) >= 12:
                if has_mirror:
                    return "Ih"
                return "I"
            if max_n == 4 and len(Cn[4]) >= 6:
                if has_mirror:
                    return "Oh"
                return "O"
            if max_n == 3 and len(Cn[3]) >= 8:
                for ele in self.elements:
                    if has_sig_d:
                        return "Td"
                    if has_mirror:
                        return "Th"
                    return "T"

            n_sig_v = len([
                ele for ele in self.elements
                if isinstance(ele, MirrorPlane) and (ele.label == "v" or ele.label == "d")
            ])
            
            if n_sig_v > max_n:
                self.LOG.warning(
                    "%i-fold rotation found, but %i sigma_v planes found\n" % (
                        max_n,
                        n_sig_v,
                    ) + 
                    "you may need to increase the maximum order of proper rotational axes that are checked"
                )
            prin_ax = Cn[max_n][0]

            n_perp = 0
            if 2 in Cn:
                for c2 in Cn[2]:
                    angle = angle_between_vectors(
                        c2.axis,
                        prin_ax.axis,
                        renormalize=False
                    )
                    if abs(angle - np.pi / 2) < rotation_tolerance:
                        n_perp += 1 

            if n_perp >= max_n:
                if has_sig_h:
                    return "D%ih" % max_n

                if n_sig_v >= max_n:
                    return "D%id" % max_n

                return "D%i" % max_n

            if has_sig_h:
                return "C%ih" % max_n

            if n_sig_v >= max_n:
                return "C%iv" % max_n

            for ele in self.elements:
                if isinstance(ele, ImproperRotation) and ele.n == 2 * max_n:
                    return "S%i" % (2 * max_n)

            return "C%i" % max_n

        if has_mirror:
            return "Cs"

        if has_inversion:
            return "Ci"

        return "C1"

    @property
    def symmetry_number(self):
        """external symmetry number"""
        n = 1
        for ele in self.elements:
            if isinstance(ele, ProperRotation):
                if ele.n > n:
                    n = ele.n
            
            elif isinstance(ele, ImproperRotation):
                if ele.n / 2 > n:
                    n = ele.n
        
        if self.name.startswith("D"):
            n *= 2
        if self.name.startswith("T"):
            n *= 4
        if self.name.startswith("O"):
            n *= 6
        if self.name.startswith("I"):
            n *= 12
        return n
    
    def equivalent_positions(self):
        """returns a list of lists of atoms that are symmetry-equivalent"""
        equivs = [set([atom]) for atom in self.geom.atoms]
        init_coords = self.geom.coords
        init_partitions = dict()
        for atom, group in zip(self.geom.atoms, self.initial_groups):
            init_partitions.setdefault(group, [])
            init_partitions[group].append(atom)
        
        init_partition_coords = {
            group: self.geom.coordinates(atoms) for group, atoms in init_partitions.items()
        }
        
        for ele in self.elements:
            if isinstance(ele, Identity):
                continue
            for group in init_partitions:
                coords = init_partition_coords[group]
                new_coords = ele.apply_operation(coords)
                
                dist = distance_matrix(coords, new_coords)
                closest_ndx = np.argmin(dist, axis=1)
                for i, (atom, ndx) in enumerate(zip(init_partitions[group], closest_ndx)):
                    if i == ndx:
                        continue
                    j = self.geom.atoms.index(atom)
                    k = self.geom.atoms.index(init_partitions[group][ndx])
                    
                    equivs[j] = {*equivs[j], init_partitions[group][ndx]}
                    equivs[k] = {*equivs[k], atom}

        # for e in equivs:
        #     for atom in sorted(e, key=lambda a: int(a.name)):
        #         print(atom)
        #     print("----------------")
        
        out = []
        for eq in equivs:
            for group in out:
                if any(atom in group for atom in eq):
                    group.extend(eq)
                    break
            else:
                out.append(list(eq))

        return out

    @property
    def optically_chiral(self):
        """is this point group optically_chiral?"""
        found_mirror = False
        found_inversion = False
        found_improper = False
        for ele in self.elements:
            if isinstance(ele, MirrorPlane):
                # S1
                found_mirror = True
            elif isinstance(ele, InversionCenter):
                # S2
                found_inversion = True
            elif isinstance(ele, ImproperRotation):
                # Sn
                found_improper = True
        
        return not (found_mirror or found_inversion or found_improper)

    def idealize_geometry(self):
        """
        adjust coordinates of self.geom to better match this point group
        also re-determines point group and symmetry elements
        """
        com = self.geom.COM()
        coords = self.geom.coords
        centered_coords = coords - com
        out = np.zeros((len(self.geom.atoms), 3))
        n_ele = len(self.elements)
        principal_axis = None

        # try moving axes of elements to more ideal positions relative to
        # a principal axis
        # for example, some point groups have elements with axes that
        # are 30, 45, 60, and 90 degrees from a principal axis
        # if we find an element close to one of these angles, we move it
        # to that angle
        # many symmetry elements will also be regularly spaced around
        # the principal_axis
        # e.g. D6h has 6 C2 axes perpendicular to the C6, with each 
        # being 30 degrees apart
        # so this looks for C2 axes (and others) that are perpendicular
        # to the principal axis and looks for C2 axes that are
        # 2 * x * pi / N (order of principal axis) apart from one of
        # the C2 axes
        # it turns out none of that really helps, so if False
        if False:
        # if principal_axis:
            principal_axis.sort(
                key=lambda ele: ele.error(
                    geom=self.geom,
                    groups=self.initial_groups,
                )
            )
            for ax in principal_axis:
                non_ideal_axes = dict()
                for ele in sorted(self.elements):
                    if not hasattr(ele, "axis"):
                        continue
                    if ax is ele:
                        continue
                    dv = ax.axis - ele.axis
                    c2 = np.linalg.norm(dv) ** 2
                    angle = np.arccos(-0.5 * (c2 - 2))
                    if np.isclose(angle, 0):
                        continue
                    for n in range(1, ax.n):
                        test_angle = np.arccos(1 / np.sqrt(n))
                        if np.isclose(angle, np.pi / 2, atol=5e-3):
                            non_ideal_axes.setdefault(np.pi / 2, [])
                            perp_axis = np.cross(ax.axis, ele.axis)
                            perp_axis = np.cross(perp_axis, ax.axis)
                            perp_axis /= np.linalg.norm(perp_axis)
                            if np.dot(perp_axis, ele.axis) < 0:
                                perp_axis *= -1
                            ele.axis = perp_axis
                            non_ideal_axes[np.pi / 2].append(ele)
                            continue
                            
                        if np.isclose(angle, test_angle, atol=5e-3):
                            non_ideal_axes.setdefault(test_angle, [])
                            non_ideal_axes[test_angle].append(ele)
                        if n == 1:
                            continue

                        test_angle = np.arccos(1. / n)
                        if any(np.isclose(a, test_angle) for a in non_ideal_axes.keys()):
                            continue
                        if np.isclose(angle, test_angle, atol=5e-3):
                            non_ideal_axes.setdefault(test_angle, [])
                            non_ideal_axes[test_angle].append(ele)

                for angle in non_ideal_axes:
                    # print(angle)
                    prop_rots = dict()
                    improp_rots = dict()
                    mirror = []
                    for ele in non_ideal_axes[angle]:
                        # print("\t", ele)
                        if isinstance(ele, MirrorPlane):
                            mirror.append(ele)
                        elif isinstance(ele, ProperRotation):
                            prop_rots.setdefault(ele.n, [])
                            prop_rots[ele.n].append(ele)
                        elif isinstance(ele, ImproperRotation):
                            improp_rots.setdefault(ele.n, [])
                            improp_rots[ele.n].append(ele)
                    
                    for n in prop_rots:
                        prop_rots[n].sort(
                            key=lambda ele: ele.error(
                                geom=self.geom,
                                groups=self.initial_groups,
                            )
                        )

                    for n in improp_rots:
                        improp_rots[n].sort(
                            key=lambda ele: ele.error(
                                geom=self.geom,
                                groups=self.initial_groups,
                            )
                        )
                    
                    mirror.sort(
                        key=lambda ele: ele.error(
                            geom=self.geom,
                            groups=self.initial_groups,
                        )
                    )
                    
                    for i in range(1, 2 * ax.n):
                        test_angle = i * np.pi / (2 * ax.n)
                        mat = rotation_matrix(test_angle, ax.axis)
                        for n in improp_rots:
                            new_v = np.matmul(mat, improp_rots[n][0].axis)
                            for ele in improp_rots[n][1:]:
                                if np.isclose(np.dot(new_v, ele.axis), 1):
                                    ele.axis = new_v
                                    break
                            else:
                                print("no C%i axis at angle" % n, test_angle)
                        
                        for n in improp_rots:
                            new_v = np.matmul(mat, improp_rots[n][0].axis)
                            for ele in improp_rots[n][1:]:
                                if np.isclose(np.dot(new_v, ele.axis), 1):
                                    ele.axis = new_v
                                
                        for mir in mirror:
                            new_v = np.matmul(mat, mirror[0].axis)
                            for ele in mirror[1:]:
                                if np.isclose(np.dot(new_v, ele.axis), 1):
                                    ele.axis = new_v
        
        # apply each operation and average the coordinates of the
        # equivalent positions
        max_n = 0
        for ele in self.elements:
            equiv = ele.equivalent_positions(coords, self.initial_groups)
            out += ele.apply_operation_without_translation(
                centered_coords[equiv]
            ) / n_ele

            if isinstance(ele, ProperRotation):
                if ele.n > max_n:
                    max_n = ele.n
        
        self.geom.coords = out + com
        self.elements = self.get_symmetry_elements(
            self.geom,
            max_rotation=max_n,
            groups=self.initial_groups,
        )

        self.determine_point_group()

    def total_error(self, return_max=False):
        tot_error = 0
        max_error = 0
        max_ele = None
        for ele in self.elements:
            error = ele.error(geom=self.geom, groups=self.initial_groups)
            tot_error += error
            if error > max_error:
                max_error = error
                max_ele = ele
        
        if return_max:
            return tot_error, max_error, max_ele
        return tot_error


"""finders are used by Geometry.find to locate atoms in a more general way"""
import sys
import inspect
from collections import deque

import numpy as np

from AaronTools import addlogger


def get_class(name):
    """returns the finder class with the given name"""
    for obj_name, obj in inspect.getmembers(sys.modules[__name__]):
        if obj_name == name and inspect.isclass(obj):
            return obj
    raise ValueError("no finder named %s in AaronTools.finders" % name)


class Finder:
    def get_matching_atoms(self, atoms, geometry=None):
        """overwrite with function that returns list(Atom) of the atoms that
        match your Finder's criteria
        geometry is an optional argument that could be used to e.g. find
        atoms a certain number of bonds"""
        pass


class BondsFrom(Finder):
    """exact number of bonds from specified atom
    avoid: bonding path cannot pass through these atoms"""
    def __init__(self, central_atom, number_of_bonds, avoid=None):
        super().__init__()

        self.central_atom = central_atom
        self.number_of_bonds = number_of_bonds
        self.avoid = avoid

    def __repr__(self):
        return "atoms %i bonds from %s" % (self.number_of_bonds, self.central_atom)

    def get_matching_atoms(self, atoms, geometry):
        """returns List(Atom) that are a certain number of bonds away from the given atom"""
        matching_atoms = []
        for atom in atoms:
            try:
                path = geometry.shortest_path(atom, self.central_atom, avoid=self.avoid)
            except LookupError:
                continue

            if len(path) - 1 == self.number_of_bonds:
                matching_atoms.append(atom)

        return matching_atoms


class WithinBondsOf(BondsFrom):
    """within a specified number of bonds from the atom"""
    def __init__(self, central_atom, number_of_bonds, **kwargs):
        super().__init__(central_atom, number_of_bonds)

    def __repr__(self):
        return "atoms within %i bonds of %s" % (self.number_of_bonds, self.central_atom)

    def get_matching_atoms(self, atoms, geometry):
        """returns List(Atom) that are a certain number of bonds away from the given atom"""
        matching_atoms = []
        stack = deque([self.central_atom])
        next_stack = deque([])
        frag = [self.central_atom]
        n_bonds = 1
        while stack and self.number_of_bonds:
            next_connected = stack.popleft()
            connected = next_connected.connected - set(frag)
            if n_bonds < self.number_of_bonds:
                next_stack.extend(connected)
            matching_atoms.extend(connected)
            frag += connected
            if not stack:
                n_bonds += 1
                stack = next_stack
                next_stack = deque([])

        return matching_atoms


class BondedTo(Finder):
    """returns all atoms that are bonded to the specified atom"""
    def __init__(self, atom):
        super().__init__()

        self.atom = atom

    def __repr__(self):
        return "atoms bonded to %s" % self.atom

    def get_matching_atoms(self, atoms, geometry=None):
        """returns list(Atom) that are within a radius of a point"""
        return [atom for atom in atoms if atom in self.atom.connected]


class WithinRadiusFromPoint(Finder):
    """within a specified radius of a point"""
    def __init__(self, point, radius):
        super().__init__()

        self.point = np.array(point)
        self.radius = radius
    
    def __repr__(self):
        return "atoms within %.2f angstroms of (%.2f, %.2f, %.2f)" % (self.radius, *self.point)

    def get_matching_atoms(self, atoms, geometry=None):
        """returns list(Atom) that are within a radius of a point"""
        keep = np.arange(0, len(atoms), dtype=int)
        coords = np.array([atom.coords for atom in atoms])
        coords -= self.point
        
        mask = np.where(coords[:, 0] < self.radius)
        coords = coords[mask]
        keep = keep[mask]
        
        mask = np.where(coords[:, 1] < self.radius)
        coords = coords[mask]
        keep = keep[mask]
        
        mask = np.where(coords[:, 2] < self.radius)
        coords = coords[mask]
        keep = keep[mask]
        
        dist = np.linalg.norm(coords, axis=1)
        mask = np.where(dist < self.radius)
        keep = keep[mask]
        
        matching_atoms = [atoms[k] for k in keep]

        return matching_atoms
    

class WithinRadiusFromAtom(Finder):
    """within a specified radius of a point"""
    def __init__(self, atom, radius):
        super().__init__()

        self.atom = atom
        self.radius = radius
    
    def __repr__(self):
        return "atoms within %.2f angstroms of %s" % (self.radius, self.atom)

    def get_matching_atoms(self, atoms, geometry=None):
        """returns list(Atom) that are within a radius of a point"""
        keep = np.arange(0, len(atoms), dtype=int)
        coords = np.array([atom.coords for atom in atoms])
        coords -= self.atom.coords
        
        mask = np.where(coords[:, 0] < self.radius)
        coords = coords[mask]
        keep = keep[mask]
        
        mask = np.where(coords[:, 1] < self.radius)
        coords = coords[mask]
        keep = keep[mask]
        
        mask = np.where(coords[:, 2] < self.radius)
        coords = coords[mask]
        keep = keep[mask]
        
        dist = np.linalg.norm(coords, axis=1)
        mask = np.where(dist < self.radius)
        keep = keep[mask]
        
        matching_atoms = [atoms[k] for k in keep]

        return matching_atoms


class NotAny(Finder):
    """atoms not matching specifiers/Finders"""
    def __init__(self, *critera, **kwargs):
        """critera can be any number of Finders and/or other atom specifiers (tags, elements, etc.)"""
        super().__init__()

        if not critera and "critera" in kwargs:
            critera = kwargs["critera"]
        if len(critera) == 1:
            if isinstance(critera[0], tuple):
                critera = critera[0]
        self.critera = critera

    def __repr__(self):
        return "not any of: %s" % ", ".join([str(x) for x in self.critera])

    def get_matching_atoms(self, atoms, geometry):
        """returns List(Atom) that do not match any of the critera"""
        unmatching_atoms = []
        for criterion in self.critera:
            try:
                unmatch = geometry.find(criterion)
                unmatching_atoms.extend(unmatch)
            
            except LookupError:
                pass

        return [atom for atom in atoms if atom not in set(unmatching_atoms)]


class AnyTransitionMetal(Finder):
    """any atoms that are transition metals"""
    def __init__(self):
        super().__init__()

    def __repr__(self):
        return "any transition metal"

    def get_matching_atoms(self, atoms, geometry=None):
        """returns List(Atom) of atoms that are metals"""
        from AaronTools.const import TMETAL
        return [atom for atom in atoms if atom.element in TMETAL]


class AnyNonTransitionMetal(NotAny):
    """any atoms that are not transition metals"""
    def __init__(self, *a, **kw):
        super().__init__(AnyTransitionMetal())

    def __repr__(self):
        return "any non-transition metal"


class HasAttribute(Finder):
    """all atoms with the specified attribute"""
    def __init__(self, attribute_name):
        super().__init__()

        self.attribute_name = attribute_name

    def __repr__(self):
        return "atoms with the '%s' attribute" % self.attribute_name

    def get_matching_atoms(self, atoms, geometry=None):
        """returns List(Atom) of atoms that have the attribute"""
        return [atom for atom in atoms if hasattr(atom, self.attribute_name)]


@addlogger
class VSEPR(Finder):
    """atoms with the specified VSEPR geometry
    see Atom.get_shape for a list of valid vsepr strings"""
    LOG = None
    def __init__(self, vsepr, cutoff=0.5):
        super().__init__()
        
        self.vsepr = vsepr
        if any(vsepr == x for x in ["triangular cupola", "heptagonal bipyramidal"]):
            self.LOG.warning(
                "triangular cupola and heptagonal bipyramidal cannot be distinguished"
            )
        self.cutoff = cutoff
    
    def __repr__(self):
        return "atoms with %s shape" % self.vsepr

    def get_matching_atoms(self, atoms, geometry=None):
        matching_atoms = []
        for atom in atoms:
            out = atom.get_vsepr()
            if out is not None:
                shape, score = atom.get_vsepr()
                if shape == self.vsepr and score < self.cutoff:
                    matching_atoms.append(atom)
        
        return matching_atoms


class BondedElements(Finder):
    """atoms bonded to the specified neighboring elements
    if match_exact=True (default), elements must match exactly 
    e.g. BondedElements('C') will find
    atoms bonded to only one carbon and nothing else"""
    def __init__(self, *args, match_exact=True, **kwargs):
        super().__init__()
        
        if not args and "elements" in kwargs:
            args = kwargs["elements"]
        self.elements = list(args)
        self.match_exact = match_exact

    def __repr__(self):
        if len(self.elements) == 0:
            return "atoms bonded to nothing"
        elif len(self.elements) == 1:
            return "atoms bonded to %s" % self.elements[0]
        else:
            return "atoms bonded to %s and %s" % (", ".join(self.elements[:-1]), self.elements[-1])
    
    def get_matching_atoms(self, atoms, geometry=None):
        matching_atoms = []
        if self.match_exact:
            ref = "".join(sorted(self.elements))
        else:
            ref = self.elements
        
        for atom in atoms:
            if self.match_exact:
                ele_list = [a.element for a in [ele for ele in atom.connected]]
                test = "".join(sorted(ele_list))
                if ref == test:
                    matching_atoms.append(atom)
            
            else:
                bonded_eles = [bonded_atom.element for bonded_atom in atom.connected]
                if all([ele in bonded_eles for ele in self.elements]):
                    matching_atoms.append(atom)
        
        return matching_atoms


class NumberOfBonds(Finder):
    """atoms with the specified number of bonds"""
    def __init__(self, num_bonds):
        super().__init__()
        
        self.num_bonds = num_bonds
    
    def __repr__(self):
        return "atoms with %i bonds" % self.num_bonds
    
    def get_matching_atoms(self, atoms, geometry=None):
        return [atom for atom in atoms if len(atom.connected) == self.num_bonds]


class ChiralCentres(Finder):
    """chiral centers
    atoms with a non-planar VSEPR geometry with all bonded groups
    being distinct
    for rings, looks for a set of unique canonical ranks for atoms that 
    are all the same number of bonds away from one atom"""
    #IUPAC spelling 
    def __init__(self, RS_only=False):
        """RS_only: bool  - if True, do not identify chiral centers that are chiral because they
                            are connected to multiple chiral fragments with the same chirality
                            this corresponds to R/S centers
                            False will include r/s centers as well
        """
        super().__init__()
        self.RS_only = RS_only

    def __repr__(self):
        return "chiral centers"

    def get_matching_atoms(self, atoms, geometry):
        from AaronTools.geometry import Geometry
        from AaronTools.symmetry import PointGroup
        # from time import perf_counter
        # 
        # start = perf_counter()
        matching_atoms = []

        # b/c they are connected to chiral fragments
        geometry.refresh_ranks()
        chiral_atoms_changed = True
        ranks = geometry.canonical_rank(break_ties=False, update=False, invariant=True)
        frags = []
        properly_shaped_atoms = []
        for atom in geometry.atoms:
            if len(atom.connected) < 3:
                continue
            vsepr, _ = atom.get_vsepr()
            if vsepr in ['trigonal planar', 't shaped', 'sqaure planar']:
                continue
            properly_shaped_atoms.append(atom)
            frags.append([])
            single_atoms = dict()
            for bonded_atom in atom.connected:
                frag = geometry.get_fragment(bonded_atom, atom, as_object=False)
                frags[-1].append(frag)
                # keep track of single atom fragments to more quickly
                # eliminate atoms that aren't chiral
                if len(frag) == 1:
                    if frag[0].element in single_atoms:
                        single_atoms[frag[0].element] += 1
                        if single_atoms[frag[0].element] >= len(atom.connected) / 2:
                            frags.pop(-1)
                            properly_shaped_atoms.pop(-1)
                            break
                    else:
                        single_atoms[frag[0].element] = 1

        # print(properly_shaped_atoms)

        # need to do multiple passes b/c sometimes atoms are chiral
        # because of other chiral centers
        k = 0
        while chiral_atoms_changed:
            chiral_atoms_changed = False
            k += 1
            #skip atoms we've already found
            for ndx, atom in enumerate(properly_shaped_atoms):
                if atom in matching_atoms:
                    continue

                neighbor_ranks = [
                    ranks[geometry.atoms.index(bonded_atom)]
                    for bonded_atom in atom.connected
                ]
                
                # first iteration should only look for centers that are chiral
                # because the fragments are different
                if k == 1 and len(atom.connected) <= 4 and all(
                    neighbor_ranks.count(rank) == 1 for rank in neighbor_ranks
                ):
                    matching_atoms.append(atom)
                    chiral_atoms_changed = True
                elif k == 1 and len(atom.connected) > 4:
                    test_geom = Geometry(
                        [atom, *atom.connected], refresh_ranks=False, refresh_connected=False
                    )
                    groups = [ranks[geometry.atoms.index(a)] for a in test_geom.atoms]
                    pg = PointGroup(test_geom, groups=groups, center=atom.coords)
                    print(pg.name)
                    if pg.name == "C1":
                        matching_atoms.append(atom)
                        chiral_atoms_changed = True
                # more iterations should only look for centers that are
                # chiral because of the presence of other chiral centers
                elif k > 1 and all(
                    neighbor_ranks.count(rank) <= len(atom.connected) / 2 for rank in neighbor_ranks
                ):
                    chiral = True
                    for i, frag1 in enumerate(frags[ndx]):
                        #get the ranks of the atoms in this fragment
                        ranks_1 = [ranks[geometry.atoms.index(atom)] for atom in frag1]
                        for frag2 in frags[ndx][:i]:
                            same = True
                            
                            ranks_2 = [ranks[geometry.atoms.index(atom)] for atom in frag2]
                            
                            if len(frag1) != len(frag2):
                                same = False
                                continue
    
                            for a, b in zip(sorted(ranks_1), sorted(ranks_2)):
                                # want correct elements
                                if a != b:
                                    same = False
                                    break
                            
                            for a, b in zip(sorted(frag1), sorted(frag2)):
                                # and other chiral atoms

                                # correct connected elements
                                for o, l in zip(
                                    sorted([aa.element for aa in a.connected]),
                                    sorted([bb.element for bb in b.connected]),
                                ):
                                    if o != l:
                                        same = False
                                        break
    
                                if a is b:
                                    break
    
                                if not self.RS_only and a in matching_atoms and b in matching_atoms:
                                    #use RMSD to see if they have the same handedness
                                    a_connected = sorted(a.connected)
                                    b_connected = sorted(b.connected)
                                    a_targets = [a] + list(a_connected)
                                    b_targets = [b] + list(b_connected)
                                    if geometry.RMSD(
                                        geometry,
                                        targets=a_targets,
                                        ref_targets=b_targets,
                                        sort=False,
                                        align=False,
                                    ) < 0.1:
                                        same = False
                                        break

                            # I'm not sure why this code was here...
                            # ring_atoms = [
                            #     bonded_atom for bonded_atom in atom.connected
                            #     if bonded_atom in frag1 and bonded_atom in frag2
                            # ]
                            # if len(ring_atoms) > 0:
                            #     #this is a ring
                            #     #look at the rank of all atoms that are n bonds away from this atom
                            #     #if the ranks are ever all different, this is a chiral center
                            #     n_bonds = 1
                            #     acceptable_nbonds = True
                            #     while acceptable_nbonds:
                            #         try:
                            #             atoms_within_nbonds = geometry.find(BondsFrom(atom, n_bonds))
                            #             nbonds_ranks = [
                            #                 ranks[geometry.atoms.index(a)] for a in atoms_within_nbonds
                            #             ]
                            #             if all(nbonds_ranks.count(r) == 1 for r in nbonds_ranks):
                            #                 same = False
                            #                 acceptable_nbonds = False
                            #             elif not self.RS_only:
                            #                 # need to find things in the ring that are chiral
                            #                 # b/c of other chiral centers
                            #                 for n, atom1 in enumerate(atoms_within_nbonds):
                            #                     for m, atom2 in enumerate(atoms_within_nbonds[n+1:]):
                            #                         p = m + n + 1
                            #                         if nbonds_ranks[n] == nbonds_ranks[p]:
                            #                             a_connected = sorted(atom1.connected)
                            #                             b_connected = sorted(atom2.connected)
                            #                             a_targets = [atom1] + list(a_connected)
                            #                             b_targets = [atom2] + list(b_connected)
                            #                             if geometry.RMSD(
                            #                                 geometry,
                            #                                 targets=a_targets,
                            #                                 ref_targets=b_targets,
                            #                                 sort=False,
                            #                                 align=False,
                            #                             ) < 0.1:
                            #                                 same = False
                            #                                 break
                            #                     if not same:
                            #                         break
                            # 
                            #             n_bonds += 1
                            #         except LookupError:
                            #             acceptable_nbonds = False
                            # 
                            #     if not same:
                            #         break
    
                            if same:
                                chiral = False
                                break

                    if chiral:
                        chiral_atoms_changed = True
                        matching_atoms.append(atom)
        
            if self.RS_only:
                break
        # stop = perf_counter()
        # print("took %.3fs" % (stop - start))
        
        return matching_atoms

#alternative spelling
ChiralCenters = ChiralCentres


class FlaggedAtoms(Finder):
    """
    atoms with a non-zero flag
    """
    # useful for finding constrained atoms
    def __repr__(self):
        return "flagged atoms"

    def get_matching_atoms(self, atoms, geometry):
        return [atom for atom in atoms if atom.flag]


class CloserTo(Finder):
    """
    atoms closer to atom1 than atom2 (based on bonds, not actual distance)
    """
    def __init__(self, atom1, atom2, include_ties=False):
        super().__init__()

        self.atom1 = atom1
        self.atom2 = atom2
        self.include_ties = include_ties

    def __repr__(self):
        return "atoms closer to %s than %s" % (self.atom1, self.atom2)

    def get_matching_atoms(self, atoms, geometry):
        matching_atoms = []
        for atom in atoms:
            if atom is self.atom1 and atom is not self.atom2:
                matching_atoms.append(atom)
                continue

            try:
                d1 = len(geometry.shortest_path(self.atom1, atom))
            except LookupError:
                d1 = False

            try:
                d2 = len(geometry.shortest_path(self.atom2, atom))
            except LookupError:
                d2 = False

            if d1 is not False and d2 is not False and d1 <= d2:
                if self.include_ties:
                    matching_atoms.append(atom)
                elif d1 < d2:
                    matching_atoms.append(atom)

            elif d1 is not False and d2 is False:
                matching_atoms.append(atom)

        return matching_atoms


class AmideCarbon(Finder):
    """
    amide carbons
    trigonal planar carbons bonded to a linear oxygen and a
    nitrogen with 3 bonds
    """
    def __repr__(self):
        return "amide carbons"

    def get_matching_atoms(self, atoms, geometry):
        matching_atoms = []

        carbons = geometry.find("C", VSEPR("trigonal planar"))
        oxygens = geometry.find("O", VSEPR("linear 1"))
        nitrogens = geometry.find("N", NumberOfBonds(3))
        for carbon in carbons:
            if (
                    any(atom in oxygens for atom in carbon.connected)
                    and any(atom in nitrogens for atom in carbon.connected)
            ):
                matching_atoms.append(carbon)

        return matching_atoms


class Bridgehead(Finder):
    """
    bridgehead atoms
    can specify ring sizes that the atoms bridge
    """
    def __init__(self, ring_sizes=None, match_exact=False):
        """
        ring_sizes  - list of int, size of rings (e.g. [6, 6] for atoms that bridge
                      two 6-membered rings)
                      not specifying yields bridgehead atoms for any ring size
        match_exact - bool, if True, return atoms only bridging the specified rings
                      if False, the ring_sizes is taken as a minimum (e.g.
                      ring_size=[6, 6], match_exact=False would also yield atoms
                      bridging three 6-membered rings or two six-membered rings and
                      a five-membered ring)
        """
        self.ring_sizes = ring_sizes
        self.match_exact = match_exact

    def __repr__(self):
        if self.ring_sizes:
            return "bridgeheads of %s-member rings" % " or ".join([str(x) for x in self.ring_sizes])
        return "bridgehead atoms"

    def get_matching_atoms(self, atoms, geometry):
        matching_atoms = []
        for atom1 in atoms:
            matching = True
            if self.ring_sizes:
                unfound_rings = list(self.ring_sizes)
            n_rings = 0
            for i, atom2 in enumerate(atom1.connected):
                for atom3 in list(atom1.connected)[:i]:
                    try:
                        path = geometry.shortest_path(atom2, atom3, avoid=atom1)
                        n_rings += 1
                        if self.ring_sizes:
                            ring_size = len(path) + 1
                            if ring_size in unfound_rings:
                                unfound_rings.remove(ring_size)
                            elif self.match_exact:
                                matching = False
                                break
                    except LookupError:
                        pass

                if not matching:
                    break

            if self.ring_sizes and not unfound_rings and matching:
                matching_atoms.append(atom1)
            elif n_rings > 1 and not self.ring_sizes:
                matching_atoms.append(atom1)

        return matching_atoms


class SpiroCenters(Finder):
    """
    atom in two different rings with no other common atoms
    """
    def __init__(self, ring_sizes=None, match_exact=False):
        """
        ring_sizes  - list of int, size of rings (e.g. [6, 6] for atoms that bridge
                      two 6-membered rings)
                      not specifying yields bridgehead atoms for any ring size
        match_exact - bool, if True, return atoms only bridging the specified rings
                      if False, the ring_sizes is taken as a minimum (e.g.
                      ring_size=[6, 6], match_exact=False would also yield atoms
                      bridging three 6-membered rings or two six-membered rings and
                      a five-membered ring)
        """
        self.ring_sizes = ring_sizes
        self.match_exact = match_exact

    def __repr__(self):
        if self.ring_sizes:
            return "atoms in different %s-member rings" % " or ".join(
                [str(x) for x in self.ring_sizes]
            )
        return "spiro atoms"

    def get_matching_atoms(self, atoms, geometry):
        matching_atoms = []
        for atom1 in atoms:
            matching = True
            if self.ring_sizes:
                unfound_rings = list(self.ring_sizes)
            n_rings = 0
            rings = []
            for i, atom2 in enumerate(atom1.connected):
                for atom3 in list(atom1.connected)[:i]:
                    try:
                        path = geometry.shortest_path(atom2, atom3, avoid=atom1)
                        for ring in rings:
                            if any(atom in path for atom in ring):
                                continue
                        rings.append(path)
                    except LookupError:
                        pass
            for i, ring in enumerate(rings):
                bad_ring = False
                for ring2 in rings[:i]:
                    if any(atom in ring for atom in ring2):
                        bad_ring = True
                        break
                if bad_ring:
                    continue
                n_rings += 1
                if self.ring_sizes:
                    ring_size = len(path) + 1
                    if ring_size in unfound_rings:
                        unfound_rings.remove(ring_size)
                    elif self.match_exact:
                        matching = False
                        break

                if not matching:
                    break

            if self.ring_sizes and not unfound_rings and matching:
                matching_atoms.append(atom1)
            elif n_rings > 1 and not self.ring_sizes:
                matching_atoms.append(atom1)

        return matching_atoms



#!/usr/bin/env python3

import argparse
import os

from sys import stdin
from warnings import warn

from AaronTools.fileIO import FileReader, read_types
from AaronTools.finders import ChiralCenters, Bridgehead, SpiroCenters, NotAny
from AaronTools.geometry import Geometry
from AaronTools.substituent import Substituent
from AaronTools.utils.utils import glob_files


changechiral_parser = argparse.ArgumentParser(
    description="change handedness of chiral centers",
    formatter_class=argparse.RawTextHelpFormatter
)

changechiral_parser.add_argument(
    "infile", metavar="input file",
    type=str,
    nargs="*",
    default=[stdin],
    help="a coordinate file"
)

changechiral_parser.add_argument(
    "-ls", "--list-chiral",
    action="store_const",
    const=True,
    default=False,
    required=False,
    dest="list_info",
    help="list information on detected chiral centers"
)

changechiral_parser.add_argument(
    "-if", "--input-format",
    type=str,
    nargs=1,
    default=None,
    choices=read_types,
    dest="input_format",
    help="file format of input - xyz is assumed if input is stdin"
)

changechiral_parser.add_argument(
    "-o", "--output-destination",
    type=str,
    default=None,
    required=False,
    metavar="output destination",
    dest="outfile",
    help="output destination\n" +
    "$i in the filename will be replaced with a number\n" +
    "if a directory is given, default is \"diastereomer-$i.xyz\" in \n" +
    "that directory\n" +
    "Default: stdout"
)

changechiral_parser.add_argument(
    "-t", "--targets",
    type=str,
    default=None,
    action="append",
    required=False,
    dest="targets",
    help="comma- or hyphen-seperated list of chiral centers to invert (1-indexed)\n" +
    "Chiral centers must have at least two fragments not in a ring\n" +
    "Detected chiral centers are atoms that:\n" +
    "    - have > 2 bonds\n" +
    "    - have a non-planar VSEPR shape\n" +
    "    - each connected fragment is distinct or is a spiro center\n" +
    "Default: change chirality of any detected chiral centers"
)

changechiral_parser.add_argument(
    "-d", "--diastereomers",
    action="store_const",
    const=True,
    default=False,
    required=False,
    dest="combos",
    help="print all diastereomers for selected chiral centers"
)

changechiral_parser.add_argument(
    "-m", "--minimize",
    action="store_true",
    default=False,
    dest="minimize",
    help="rotate substituents to mitigate steric clashing",
)


args = changechiral_parser.parse_args()

s = ""

for infile in glob_files(args.infile, parser=changechiral_parser):
    if isinstance(infile, str):
        if args.input_format is not None:
            f = FileReader((infile, args.input_format[0], infile))
        else:
            f = FileReader(infile)
    else:
        if args.input_format is not None:
            f = FileReader(("from stdin", args.input_format[0], stdin))
        else:
            f = FileReader(("from stdin", "xyz", stdin))

    geom = Geometry(f)

    target_list = []
    if args.targets is None:
        try:
            chiral_centers = geom.find(ChiralCenters())
            spiro_centers = geom.find(SpiroCenters(), chiral_centers)
            bridge_centers = geom.find(chiral_centers, Bridgehead(), NotAny(spiro_centers))
            target_list = [t for t in chiral_centers if t not in bridge_centers]
        except LookupError as e:
            warn(str(e))
    else:
        for targ in args.targets:
            target_list.extend(geom.find(targ))

    if args.list_info:
        if len(args.infile) > 1:
            s += "%s\n" % infile
        s += "Target\tElement\n"
        for targ in target_list:
            s += "%-2s\t%-2s\n" % (targ.name, targ.element)
        if infile is not args.infile[-1]:
            s += "\n"
        continue

    geom.substituents = []
    if args.combos:
        # this stuff is copy-pasted from makeConf, so it's a bit overkill
        # for getting all diastereomers, as each chiral center can only
        # have 2 options instead of the random number of rotamers
        # a substituent can have
        diastereomers = Geometry.get_diastereomers(geom, minimize=args.minimize)
        
        for i, diastereomer in enumerate(diastereomers):
            if args.outfile is None:
                s += diastereomer.write(outfile=False)
                s += "\n"
            else:
                if os.path.isdir(os.path.expanduser(args.outfile)):
                    outfile = os.path.join(
                        os.path.expanduser(args.outfile),
                        "diastereomer-%i.xyz" % (i + 1)
                    )
    
                else:
                    outfile = args.outfile.replace("$i", str(i + 1))
    
                diastereomer.write(outfile=outfile, append="$i" not in args.outfile)

    else:
        for targ in target_list:
            geom.change_chirality(targ)

        if args.minimize:
            geom.minimize_sub_torsion(increment=15)

        if args.outfile is None:
            s += geom.write(outfile=False)
            s += "\n"
        else:
            geom.write(outfile=args.outfile)


if args.outfile is None or args.list_info:
    print(s[:-1])



#!/usr/bin/env python3
import ast
import dis
import importlib
import inspect
import os
import re
from importlib import resources
from pprint import pprint

from AaronTools import getlogger

LOG = getlogger()


def main(args):
    modules = [
        r[:-3] for r in resources.contents(args.package) if r.endswith(".py")
    ]

    obj = None
    class_name, function_name = None, None
    if os.path.isfile(args.name):
        with open(args.name) as f:
            obj = compile(f.read(), filename=args.name, mode="exec")
    else:
        match = re.match("(\w+)\.(\w+)", args.name)
        if match is not None:
            class_name, function_name = match.groups()
        else:
            match = re.match("(\w+)", args.name)
            if match is not None:
                class_name, function_name = match.group(1), None
        for module in modules:
            module_name = module
            module = "{}.{}".format(args.package, module_name)
            try:
                module = importlib.import_module(module)
            except Exception:
                continue
            all_classes = inspect.getmembers(module, inspect.isclass)
            for c in all_classes:
                if class_name == c[0]:
                    obj = c[1]
                    break
            else:
                continue
            break
        if function_name is not None:
            obj = getattr(obj, function_name)
    if obj is None and class_name is not None:
        if function_name is None:
            LOG.error("Cannot find %s in %s", class_name, args.package)
        else:
            LOG.error(
                "Cannot find %s.%s in %s",
                class_name,
                function_name,
                args.package,
            )
        exit(1)
    elif obj is None:
        LOG.error("Cannot load %s", args.name)
        exit(1)
    for citation in set(get_citations(obj, args.name)):
        print(*citation)


def get_citations(obj, obj_name, done=None, citations=None):
    if done is None:
        done = set([])
    if citations is None:
        citations = []
    if obj in done:
        return citations
    done.add(obj)

    try:
        instructions = [inst for inst in dis.get_instructions(obj)]
    except TypeError:
        return citations
    names = {}
    methods = {}
    add_obj = set([])
    for i, inst in enumerate(instructions):
        if (
            inst.argval == "CITATION"
            and instructions[i - 1].opname == "LOAD_CONST"
        ):
            names.setdefault(inst.argval, [])
            names[inst.argval].append((obj_name, instructions[i - 1].argval))
        elif inst.opname == "STORE_NAME":
            names.setdefault(inst.argval, [])
            for prev in reversed(instructions[:i]):
                if prev.opname == "STORE_NAME":
                    break
                if prev.opname == "POP_TOP":
                    break
                if "JUMP" in prev.opname:
                    break
                if prev.opname == "IMPORT_FROM":
                    names[inst.argval].append(prev.argval)
                if prev.opname == "IMPORT_NAME":
                    names[inst.argval].append(prev.argval)
                    break
                if prev.opname == "LOAD_NAME":
                    names[inst.argval].append(prev.argval)
            names[inst.argval].reverse()
        elif (
            inst.opname in ["LOAD_ATTR", "LOAD_METHOD"]
            and instructions[i - 1].argval == "self"
        ):
            add_obj.add(".".join(obj_name.split(".")[:-1] + [inst.argval]))
        elif inst.opname == "LOAD_METHOD":
            methods[inst.argval] = instructions[i - 1].argval

    if "CITATION" in names:
        citations += names["CITATION"]
    for rm in get_recurse_methods(names, methods, add_obj=add_obj):
        citations = get_citations(*rm, done=done, citations=citations)
    return citations


def get_recurse_methods(name_dict, method_dict, add_obj=None):
    if add_obj is None:
        recurse_methods = set([])
    else:
        recurse_methods = add_obj
    for method, name_key in method_dict.items():
        if name_key not in name_dict:
            continue
        name_list = name_dict[name_key]
        tmp = []
        for name in name_list:
            if name in name_dict:
                for i in name_dict[name]:
                    if i in tmp:
                        continue
                    tmp.append(i)
            elif name not in tmp:
                tmp.append(name)
        recurse_methods.add(".".join(tmp + [method]))

    rv = set([])
    for rm in recurse_methods:
        module_name = rm
        obj = None
        while True:
            try:
                obj = importlib.import_module(module_name)
                break
            except ModuleNotFoundError:
                module_name = module_name.rsplit(".", maxsplit=1)[0]
        remainder = rm.replace(module_name + ".", "").split(".")
        if len(remainder) == 2:
            class_name, function_name = remainder[0], remainder[1]
        elif len(remainder) == 1:
            class_name, function_name = remainder[0], None
        elif len(remainder) == 0:
            class_name, function_name = None, None
        else:
            raise Exception

        all_classes = inspect.getmembers(obj, inspect.isclass)
        for c in all_classes:
            if c[0] == class_name:
                obj = c[1]
                break
        init_method = getattr(obj, "__init__")
        rv.add((init_method, "{}.{}.__init__".format(module_name, class_name)))
        if function_name is not None and hasattr(obj, function_name):
            obj = getattr(obj, function_name)
        rv.add((obj, rm))
    return rv


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("name")
    parser.add_argument("-p", "--package", default="AaronTools")
    args = parser.parse_args()
    main(args)



#!/usr/bin/env python3

import argparse
import sys

from AaronTools.fileIO import FileReader, read_types
from AaronTools.geometry import Geometry
from AaronTools.substituent import Substituent
from AaronTools.utils.utils import get_filename, glob_files

substitute_parser = argparse.ArgumentParser(
    description="replace an atom or substituent with another",
    formatter_class=argparse.RawTextHelpFormatter,
)
substitute_parser.add_argument(
    "infile",
    metavar="input file",
    type=str,
    nargs="*",
    default=[sys.stdin],
    help="a coordinate file",
)

substitute_parser.add_argument(
    "-ls",
    "--list",
    action="store_const",
    const=True,
    default=False,
    required=False,
    dest="list_avail",
    help="list available substituents",
)

substitute_parser.add_argument(
    "-if",
    "--input-format",
    type=str,
    default=None,
    choices=read_types,
    dest="input_format",
    help="file format of input - xyz is assumed if input is stdin",
)

substitute_parser.add_argument(
    "-s",
    "--substitute",
    metavar="n=substituent",
    type=str,
    action="append",
    default=None,
    required=False,
    dest="substitutions",
    help="substitution instructions \n"
    + "n is the 1-indexed position of the starting position of the\n"
    + "substituent you are replacing\n"
    + "a substituent name prefixed by iupac: or smiles: (e.g. iupac:acetyl\n"
    + "or smiles:O=[N.]=O) will create the substituent from the\n"
    + "corresponding identifier",
)

substitute_parser.add_argument(
    "-m",
    "--minimize",
    action="store_const",
    const=True,
    default=False,
    required=False,
    dest="mini",
    help="rotate substituents to try to minimize LJ energy",
)

substitute_parser.add_argument(
    "-o",
    "--output",
    type=str,
    default=False,
    required=False,
    metavar="output destination",
    dest="outfile",
    help="output destination\n" +
    "$INFILE will be replaced with the name of the input file\n" +
    "Default: stdout"
)

args = substitute_parser.parse_args()

if args.list_avail:
    s = ""
    for i, name in enumerate(sorted(Substituent.list())):
        s += "%-20s" % name
        # if (i + 1) % 3 == 0:
        if (i + 1) % 1 == 0:
            s += "\n"

    print(s.strip())
    sys.exit(0)

for infile in glob_files(args.infile, parser=substitute_parser):
    if isinstance(infile, str):
        if args.input_format is not None:
            f = FileReader((infile, args.input_format, infile))
        else:
            f = FileReader(infile)
    else:
        if args.input_format is not None:
            f = FileReader(("from stdin", args.input_format, infile))
        else:
            f = FileReader(("from stdin", "xyz", infile))

    geom = Geometry(f)

    target_list = []
    for sub in args.substitutions:
        ndx_targets = sub.split("=")[0]
        target_list.append(geom.find(ndx_targets))

    for i, sub in enumerate(args.substitutions):
        ndx_target = target_list[i]
        sub_name = "=".join(sub.split("=")[1:])

        for target in ndx_target:
            if sub_name.lower().startswith("iupac:"):
                sub_name = ":".join(sub_name.split(":")[1:])
                sub = Substituent.from_string(sub_name, form="iupac")
            elif sub_name.lower().startswith("smiles:"):
                sub_name = ":".join(sub_name.split(":")[1:])
                sub = Substituent.from_string(sub_name, form="smiles")
            else:
                sub = Substituent(sub_name)

            # replace old substituent with new substituent
            geom.substitute(sub, target, minimize=args.mini)
            geom.refresh_connected()

    if args.outfile:
        outfile = args.outfile
        if "INFILE" in outfile:
            outfile = outfile.replace("$INFILE", get_filename(infile))
        geom.write(
            append=True,
            outfile=outfile,
        )
    else:
        print(geom.write(outfile=False))



#! /usr/bin/env python3
import os
import unittest

from AaronTools.test import TestWithTimer, prefix, rmsd_tol, validate
from AaronTools.utils import utils


class TestUtils(TestWithTimer):
    def test_same_cycle(self):
        graph = [[1, 2], [0, 2], [0, 1, 3], [2]]
        self.assertFalse(utils.same_cycle(graph, 2, 3))
        self.assertFalse(utils.same_cycle(graph, 0, 3))
        self.assertTrue(utils.same_cycle(graph, 0, 1))

        graph = [[1, 2, 3], [0, 2], [0, 1], [0, 4], [3]]
        self.assertFalse(utils.same_cycle(graph, 2, 3))
        self.assertFalse(utils.same_cycle(graph, 0, 4))
        self.assertTrue(utils.same_cycle(graph, 1, 0))
        self.assertTrue(utils.same_cycle(graph, 2, 0))


def suite():
    suite = unittest.TestSuite()
    suite.addTest(TestUtils("test_same_cycle"))
    return suite


if __name__ == "__main__":
    # for running specific tests, change test name in suite()
    runner = unittest.TextTestRunner()
    runner.run(suite())



#! /usr/bin/env python3
from AaronTools.fileIO import FileReader


def main(args):
    fr = FileReader(args.filename, just_geom=False)
    if args.show:
        s = "frequency\t"
        for x in args.show:
            if x == "vector":
                continue
            s += "%s\t" % x
        print(s)
    freq = fr.other["frequency"]
    if not any((args.fundamentals, args.overtones, args.combinations)):
        for i, data in enumerate(sorted(freq.data, key=lambda x: x.frequency)):
            if args.type == "neg" and data.frequency > 0:
                continue
            if args.type == "pos" and data.frequency < 0:
                continue
            if isinstance(args.type, int) and i + 1 > args.type:
                break
            s = "%9.4f\t" % data.frequency
            for x in args.show:
                if x == "vector":
                    continue
                val = getattr(data, x)
                if isinstance(val, float):
                    s += "%9.4f\t" % val
                else:
                    s += "%s\t" % str(val)
            
            print(s)
            
            if "vector" in args.show:
                print(data.vector)
                print()
    if freq.anharm_data:
        for i, data in enumerate(sorted(freq.anharm_data, key=lambda x: x.frequency)):
            if args.fundamentals:
                s = "%9.4f\t" % data.frequency
                for x in args.show:
                    if not hasattr(data, x):
                        continue
                    val = getattr(data, x)
                    if isinstance(val, float):
                        s += "%9.4f\t" % val
                    else:
                        s += "%s\t" % str(val)
                
                print(s)
            
            if args.overtones:
                for k, overtone in enumerate(data.overtones):
                    s = "%i x %9.4f = %9.4f\t" % (
                        (k + 2), data.frequency, overtone.frequency
                    )
                    for x in args.show:
                        if not hasattr(overtone, x):
                            continue
                        val = getattr(overtone, x)
                        if isinstance(val, float):
                            s += "%9.4f\t" % val
                        else:
                            s += "%s\t" % str(val)
                    
                    print(s)
            
            if args.combinations:
                for key in data.combinations:
                    for combo in data.combinations[key]:
                        s = "%9.4f + %9.4f = %9.4f\t" % (
                            data.frequency,
                            freq.anharm_data[key].frequency,
                            combo.frequency,
                        )
                        for x in args.show:
                            if not hasattr(combo, x):
                                continue
                            val = getattr(combo, x)
                            if isinstance(val, float):
                                s += "%9.4f\t" % val
                            else:
                                s += "%s\t" % str(val)

                        print(s)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="Prints frequencies from computational output file",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "filename", help="Completed QM output file with frequency info"
    )
    parser.add_argument(
        "--type",
        "-t",
        type=str,
        nargs="?",
        help="Types of frequencies to print (defaults to `all`. Allowed values: `all`, `neg[ative]`, `pos[itive]`, or `x` to print the first int(x) modes)",
        default="all",
    )
    parser.add_argument(
        "--show",
        "-s",
        type=str,
        nargs="*",
        help="Specify what additional information to show\n"
        "Some info may not be available for certain file formats",
        choices=["intensity", "vector", "forcek",
        "symmetry", "delta_anh", "harmonic_frequency",
        "harmonic_intensity", "rotation", "raman_activity"],
        default=[],
    )
    parser.add_argument(
        "--fundamentals",
        "-f",
        action="store_true",
        default=False,
        dest="fundamentals",
        help="print anharmonic fundamental frequencies for files with anharmonic data",
    )
    parser.add_argument(
        "--overtone-bands",
        "-ob",
        action="store_true",
        default=False,
        dest="overtones",
        help="print overtone frequencies for files with anharmonic data",
    )
    parser.add_argument(
        "--combination-bands",
        "-cb",
        action="store_true",
        default=False,
        dest="combinations",
        help="print combination frequencies for files with anharmonic data",
    )
    args = parser.parse_args()
    try:
        args.type = int(args.type)
    except ValueError:
        args.type = args.type.lower()
        if args.type not in ["all", "neg", "negative", "pos", "positive"]:
            parser.print_help()
            exit(1)
        elif args.type in ["neg", "negative"]:
            args.type = "neg"
        elif args.type in ["pos", "positive"]:
            args.type = "pos"
    main(args)



#!/usr/bin/env python3

import sys
import argparse

from AaronTools.geometry import Geometry
from AaronTools.fileIO import FileReader, read_types
from AaronTools.utils.utils import get_filename, glob_files

xyz_parser = argparse.ArgumentParser(
    description="print structure in xyz format",
    formatter_class=argparse.RawTextHelpFormatter
)

xyz_parser.add_argument(
    "infile", metavar="input file",
    type=str,
    nargs="*",
    default=[sys.stdin],
    help="a coordinate file"
)

xyz_parser.add_argument(
    "-o", "--output",
    type=str,
    default=False,
    required=False,
    dest="outfile",
    help="output destination\n" +
    "$INFILE will be replaced with the name of the input file\n" +
    "Default: stdout"
)

xyz_parser.add_argument(
    "-if", "--input-format",
    type=str,
    default=None,
    dest="input_format",
    choices=read_types,
    help="file format of input - xyz is assumed if input is stdin"
)

xyz_parser.add_argument(
    "-c", "--comment",
    type=str,
    default=None,
    required=False,
    dest="comment",
    help="comment line"
)

xyz_parser.add_argument(
    "-a", "--append",
    action="store_true",
    default=False,
    required=False,
    dest="append",
    help="append structures to output file if it already exists\nDefault: false"
)

args = xyz_parser.parse_args()

for f in glob_files(args.infile, parser=xyz_parser):
    if isinstance(f, str):
        if args.input_format is not None:
            infile = FileReader((f, args.input_format, None))
        else:
            infile = FileReader(f)
    else:
        if args.input_format is not None:
            infile = FileReader(("from stdin", args.input_format, f))
        else:
            if len(sys.argv) >= 1:
                infile = FileReader(("from stdin", "xyz", f))

    geom = Geometry(infile, refresh_connected=False, refresh_ranks=False)
    if args.comment:
        geom.comment = args.comment
    else:
        geom.comment = f


    if not args.outfile:
        print(geom.write(outfile=False))
    else:
        outfile = args.outfile
        if "$INFILE" in outfile:
            outfile = outfile.replace("$INFILE", get_filename(f))
        geom.write(append=args.append, outfile=outfile)



"""
used for specifying basis information for a Theory()
"""

import os
import re
from warnings import warn

from AaronTools import addlogger
from AaronTools.const import ELEMENTS
from AaronTools.finders import (
    AnyNonTransitionMetal,
    AnyTransitionMetal,
    NotAny,
)
from AaronTools.theory import (
    GAUSSIAN_GEN_BASIS,
    GAUSSIAN_GEN_ECP,
    GAUSSIAN_ROUTE,
    ORCA_BLOCKS,
    ORCA_ROUTE,
    PSI4_BEFORE_GEOM,
    QCHEM_REM,
    QCHEM_SETTINGS,
)


@addlogger
class Basis:
    """
    has attributes:
    name          - same as initialization keyword
    elements      - same as initialization keyword
    aux_type      - same as initialization keyword
    elements      - list of element symbols for elements this basis applies to
                    updated with Basis.refresh_elements
                    Basis.refresh_elements is called when writing an input file
    ele_selection - list of finders used to determine which elements this basis applies to
    not_anys      - list of finders used to determine which elements this basis does not apply to
    """

    LOG = None

    default_elements = [AnyTransitionMetal(), AnyNonTransitionMetal()]

    def __init__(self, name, elements=None, aux_type=None, user_defined=False):
        """
        name         -   basis set base name (e.g. 6-31G)
        elements     -   list of element symbols or finders to determine the basis set applies to
                         elements may also be 'tm' or 'all' to indicate any transition metal and
                         all elements, respectively
                         elements may start with '!' to exclude that element from the basis
                         for example, elements='!H' will apply to default elements, minus H
        aux_type     -   str - ORCA: one of BasisSet.ORCA_AUX; Psi4: one of BasisSet.PSI4_AUX
        user_defined -   path to file containing basis info from basissetexchange.org or similar
                         False for builtin basis sets
        """
        self.name = name

        if elements is None:
            self.elements = []
            self.ele_selection = self.default_elements
            self.not_anys = []
        else:
            # a list of elements or other identifiers was given
            # if it's an element with a ! in front, add that element to not_anys
            # otherwise, add the appropriate thing to ele_selection
            if not hasattr(elements, "__iter__") or isinstance(elements, str):
                elements = [elements]

            self.elements = elements
            ele_selection = []
            not_anys = []
            for ele in elements:
                not_any = False
                if isinstance(ele, str) and ele.startswith("!"):
                    ele = ele.lstrip("!")
                    not_any = True

                if ele.lower() == "all":
                    if not_any:
                        not_anys.append(AnyTransitionMetal())
                        not_anys.append(AnyNonTransitionMetal())
                    else:
                        ele_selection.append(AnyTransitionMetal())
                        ele_selection.append(AnyNonTransitionMetal())
                elif ele.lower() == "tm" and ele != "Tm":
                    if not_any:
                        ele_selection.append(AnyNonTransitionMetal())
                    else:
                        ele_selection.append(AnyTransitionMetal())
                elif ele.lower() == "!tm" and ele != "!Tm":
                    if not_any:
                        ele_selection.append(AnyNonTransitionMetal())
                    else:
                        ele_selection.append(AnyNonTransitionMetal())
                elif isinstance(ele, str) and ele in ELEMENTS:
                    if not_any:
                        not_anys.append(ele)
                    else:
                        ele_selection.append(ele)
                else:
                    warn("element not known: %s" % repr(ele))

            if not ele_selection and not_anys:
                # if only not_anys were given, fall back to the default elements
                ele_selection = self.default_elements

            self.ele_selection = ele_selection
            self.not_anys = not_anys

        self.aux_type = aux_type
        self.user_defined = user_defined

    def __repr__(self):
        return "%s(%s)%s" % (
            self.name,
            " ".join(self.elements),
            "" if not self.aux_type else "/%s" % self.aux_type
        )

    def __lt__(self, other):
        if self.name < other.name:
            return True
        elif self.name == other.name and self.elements and other.elements:
            return self.elements[0] < other.elements[0]
        return False

    def __eq__(self, other):
        if self.__class__ is not other.__class__:
            return False

        if (
            self.get_gaussian(self.name).lower()
            != other.get_gaussian(other.name).lower()
        ):
            return False

        if self.aux_type != other.aux_type:
            return False

        for obj, obj2 in zip([self, other], [other, self]):
            for finder in obj.ele_selection:
                if isinstance(finder, str):
                    if finder not in obj2.ele_selection:
                        return False
                else:
                    for finder2 in obj2.ele_selection:
                        if repr(finder) == repr(finder2):
                            break
                    else:
                        return False

        return True

    def copy(self):
        new_dict = dict()
        for key, value in self.__dict__.items():
            try:
                new_dict[key] = value.copy()
            except AttributeError:
                new_dict[key] = value
                # ignore chimerax objects so seqcrow doesn't print a
                # warning when a geometry is copied
                if "chimerax" in value.__class__.__module__:
                    continue
                if value.__class__.__module__ != "builtins":
                    self.LOG.warning(
                        "No copy method for {}: in-place changes may occur".format(
                            type(value)
                        )
                    )
        
        return self.__class__(**new_dict)

    def refresh_elements(self, geometry):
        """sets self's elements for the geometry"""
        atoms = geometry.find(self.ele_selection, NotAny(*self.not_anys))
        elements = set([atom.element for atom in atoms])
        self.elements = sorted(elements)

    @staticmethod
    def sanity_check_basis(name, program):
        import os.path
        from difflib import SequenceMatcher as seqmatch
        from re import IGNORECASE, match

        from AaronTools.const import AARONTOOLS
        from numpy import argsort, loadtxt

        warning = None

        if program.lower() == "gaussian":
            valid = loadtxt(
                os.path.join(
                    AARONTOOLS, "theory", "valid_basis_sets", "gaussian.txt"
                ),
                dtype=str,
            )
        elif program.lower() == "orca":
            valid = loadtxt(
                os.path.join(
                    AARONTOOLS, "theory", "valid_basis_sets", "orca.txt"
                ),
                dtype=str,
            )
        elif program.lower() == "psi4":
            valid = loadtxt(
                os.path.join(
                    AARONTOOLS, "theory", "valid_basis_sets", "psi4.txt"
                ),
                dtype=str,
            )
        elif program.lower() == "qchem":
            valid = loadtxt(
                os.path.join(
                    AARONTOOLS, "theory", "valid_basis_sets", "qchem.txt"
                ),
                dtype=str,
            )
        else:
            raise NotImplementedError(
                "cannot validate basis names for %s" % program
            )

        if not any(
            # need to escape () b/c they aren't capturing groups, it's ccsd(t) or something
            match(
                "%s$"
                % (basis.replace("(", "\(").replace(")", "\)"))
                .replace("*", "\*")
                .replace("+", "\+"),
                name,
                flags=IGNORECASE,
            )
            for basis in valid
        ):
            warning = (
                "basis '%s' may not be available in %s\n" % (name, program)
                + "if this is incorrect, please submit a bug report at https://github.com/QChASM/AaronTools.py/issues"
            )

            # try to suggest alternatives that have similar names
            simm = [
                seqmatch(
                    lambda x: x in "-_()/", name.upper(), test_basis.upper()
                ).ratio()
                for test_basis in valid
            ]
            ndx = argsort(simm)[-5:][::-1]
            warning += "\npossible misspelling of:\n"
            warning += "\n".join([valid[i] for i in ndx])

        return warning

    @staticmethod
    def get_gaussian(name):
        """
        returns the Gaussian09/16 name of the basis set
        currently just removes the hyphen from the Karlsruhe def2 ones
        """
        if name.startswith("def2-"):
            return name.replace("def2-", "def2", 1)

        return name

    @staticmethod
    def get_orca(name):
        """
        returns the ORCA name of the basis set
        currently just adds hyphen to Karlsruhe basis if it isn't there
        """
        if name.startswith("def2") and not re.match("def2(?:-|$)", name):
            return name.replace("def2", "def2-", 1)

        return name

    @staticmethod
    def get_psi4(name):
        """
        returns the Psi4 name of the basis set
        currently just adds hyphen to Karlsruhe basis if it isn't there
        """
        if name.startswith("def2") and not name.startswith("def2-"):
            return name.replace("def2", "def2-", 1)

        # pople basis sets don't have commas
        # e.g. 6-31G(d,p) -> 6-31G(d_p)
        if name.startswith("6-31") or name.startswith("3-21") or name.lower().startswith("sto"):
            name = name.replace(",", "_")

        return name

    @staticmethod
    def get_qchem(name):
        """
        returns the Psi4 name of the basis set
        currently just adds hyphen to Karlsruhe basis if it isn't there
        """
        if name.startswith("def2") and not name.startswith("def2-"):
            return name.replace("def2", "def2-", 1)

        return name


class ECP(Basis):
    """ECP - aux info will be ignored"""

    default_elements = (AnyTransitionMetal(), )

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)

    def __eq__(self, other):
        if not isinstance(other, ECP):
            return False

        return super().__eq__(other)

    @staticmethod
    def sanity_check_basis(name, program):
        import os.path
        from difflib import SequenceMatcher as seqmatch
        from re import IGNORECASE, match

        from AaronTools.const import AARONTOOLS
        from numpy import argsort, loadtxt

        warning = None

        if program.lower() == "gaussian":
            valid = loadtxt(
                os.path.join(
                    AARONTOOLS,
                    "theory",
                    "valid_basis_sets",
                    "gaussian_ecp.txt",
                ),
                dtype=str,
            )
        elif program.lower() == "orca":
            valid = loadtxt(
                os.path.join(
                    AARONTOOLS, "theory", "valid_basis_sets", "orca_ecp.txt"
                ),
                dtype=str,
            )
        elif program.lower() == "qchem":
            valid = loadtxt(
                os.path.join(
                    AARONTOOLS, "theory", "valid_basis_sets", "qchem_ecp.txt"
                ),
                dtype=str,
            )
        else:
            raise NotImplementedError(
                "cannot validate ECP names for %s" % program
            )

        if not any(
            # need to escape () b/c they aren't capturing groups, it's ccsd(t) or something
            match(
                "%s$"
                % (basis.replace("(", "\(").replace(")", "\)"))
                .replace("*", "\*")
                .replace("+", "\+"),
                name,
                flags=IGNORECASE,
            )
            for basis in valid
        ):
            warning = (
                "ECP '%s' may not be available in %s\n" % (name, program)
                + "if this is incorrect, please submit a bug report at https://github.com/QChASM/AaronTools.py/issues"
            )

            # try to suggest alternatives that have similar names
            simm = [
                seqmatch(
                    lambda x: x in "-_()/", name.upper(), test_basis.upper()
                ).ratio()
                for test_basis in valid
            ]
            ndx = argsort(simm)[-5:][::-1]
            warning += "\npossible misspelling of:\n"
            warning += "\n".join([valid[i] for i in ndx])

        return warning


@addlogger
class BasisSet:
    """used to more easily get basis set info for writing input files"""

    LOG = None

    ORCA_AUX = ["C", "J", "JK", "CABS", "OptRI CABS"]
    PSI4_AUX = [
        "JK", "RI", "DF SCF", "DF SAPT", "DF GUESS", "DF SAD", "DF MP2",
        "DF CC", "DF DCT", "DF MCSCF", "DF ELST"
    ]
    QCHEM_AUX = ["RI", "J", "K", "corr"]

    def __init__(self, basis=None, ecp=None):
        """
        basis: list(Basis), Basis, str, or None
        ecp: list(ECP) or None
        """
        if isinstance(basis, str):
            basis = self.parse_basis_str(basis, cls=Basis)
        elif isinstance(basis, Basis):
            basis = [basis]
        elif isinstance(basis, BasisSet):
            if ecp is None:
                ecp = basis.ecp
            basis = basis.basis

        if isinstance(ecp, str):
            if ecp.split():
                ecp = self.parse_basis_str(ecp, cls=ECP)
            else:
                ecp = [ECP(ecp)]
        elif isinstance(ecp, ECP):
            ecp = [ecp]
        elif isinstance(ecp, BasisSet):
            ecp = ecp.ecp

        self.basis = basis
        self.ecp = ecp

    @property
    def elements_in_basis(self):
        """returns a list of elements in self's basis"""
        elements = []
        if self.basis is not None:
            for basis in self.basis:
                elements.extend(basis.elements)

        return elements

    @staticmethod
    def parse_basis_str(basis_str, cls=Basis):
        """
        parse basis set specification string and returns list(cls)
        cls should be Basis or ECP (or subclasses of these)
        basis info should have:
            - a list of elements before basis set name (e.g. C H N O)
                - other element keywords are 'tm' (transition metals) and 'all' (all elements)
                - can also put "!" before an element to exclude it from the basis set
            - auxilliary type before basis name (e.g. auxilliary C)
            - basis set name
            - path to basis set file right after basis set name if the basis is not builtin
                - path cannot contain spaces
        Example:
            "!H !tm def2-SVPD /home/CoolUser/basis_sets/def2svpd.gbs H def2-SVP Ir SDD
        """
        info = basis_str.split()
        i = 0
        basis_sets = []
        elements = []
        aux_type = None
        user_defined = False
        while i < len(info):
            if info[i].lstrip("!") in ELEMENTS or any(
                info[i].lower().lower().lstrip("!") == x for x in ["all", "tm"]
            ):
                elements.append(info[i])
            elif info[i].lower().startswith("aux"):
                try:
                    aux_type = info[i + 1]
                    i += 1
                    if any(aux_type.lower() == x for x in ["df", "optri"]):
                        aux_type += " %s" % info[i + 1]
                        i += 1
                except:
                    raise RuntimeError(
                        'error while parsing basis set string: %s\nfound "aux"'
                        + ", but no auxilliary type followed" % basis_str
                    )
            else:
                basis_name = info[i]
                try:
                    # TODO: allow spaces in paths
                    if (
                        # os thinks I have a file named "aux" somewhere on my computer
                        # I don't see it, but basis file names cannot start with 'aux'
                        os.path.exists(info[i + 1])
                        and not info[i + 1].lower().startswith("aux")
                    ) or os.sep in info[i + 1]:
                        user_defined = info[i + 1]
                        i += 1
                except IndexError:
                    pass

                if not elements:
                    elements = None

                basis_sets.append(
                    cls(
                        basis_name,
                        elements=elements,
                        aux_type=aux_type,
                        user_defined=user_defined,
                    )
                )
                elements = []
                aux_type = None
                user_defined = False
            i += 1
        return basis_sets

    def __eq__(self, other):
        if self.__class__ is not other.__class__:
            return False
        if self.basis and other.basis:
            if len(self.basis) != len(other.basis):
                return False
            for b1, b2 in zip(sorted(self.basis), sorted(other.basis)):
                if b1 != b2:
                    return False
        else:
            if self.basis != other.basis:
                return False

        if self.ecp and other.ecp:
            if len(self.ecp) != len(other.ecp):
                return False
            for b1, b2 in zip(sorted(self.ecp), sorted(other.ecp)):
                if b1 != b2:
                    return False
        else:
            if bool(self.ecp) != bool(other.ecp):
                return False

        return True
    
    def copy(self):
        new_dict = dict()
        for key, value in self.__dict__.items():
            try:
                new_dict[key] = value.copy()
            except AttributeError:
                new_dict[key] = value
                # ignore chimerax objects so seqcrow doesn't print a
                # warning when a geometry is copied
                if "chimerax" in value.__class__.__module__:
                    continue
                if value.__class__.__module__ != "builtins":
                    self.LOG.warning(
                        "No copy method for {}: in-place changes may occur".format(
                            type(value)
                        )
                    )
        
        return self.__class__(**new_dict)

    def add_ecp(self, ecp):
        """add ecp to this BasisSet
        ecp - ECP"""
        if self.ecp is None:
            self.ecp = []

        self.ecp.append(ecp)

    def add_basis(self, basis):
        """add basis to this BasisSet
        basis - Basis"""
        if self.basis is None:
            self.basis = []

        self.basis.append(basis)

    def refresh_elements(self, geometry):
        """evaluate element specifications for each basis and ecp to
        make them compatible with the supplied geometry"""
        if self.basis is not None:
            for basis in self.basis:
                basis.refresh_elements(geometry)

        if self.ecp is not None:
            for ecp in self.ecp:
                ecp.refresh_elements(geometry)

    def get_gaussian_basis_info(self):
        """returns dict used by get_gaussian_header/footer with basis info"""
        info = {}
        warnings = []

        if self.basis is not None:
            # check if we need to use gen or genecp:
            #    -a basis set is user-defined (stored in an external file e.g. from the BSE)
            #    -multiple basis sets
            #    -an ecp
            if (
                all([basis == self.basis[0] for basis in self.basis])
                and not self.basis[0].user_defined
                and self.ecp is None
            ):
                basis_name = Basis.get_gaussian(self.basis[0].name)
                warning = self.basis[0].sanity_check_basis(
                    basis_name, "gaussian"
                )
                if warning:
                    warnings.append(warning)
                info[GAUSSIAN_ROUTE] = "/%s" % basis_name
            else:
                if self.ecp is None or all(
                    not ecp.elements for ecp in self.ecp
                ):
                    info[GAUSSIAN_ROUTE] = "/gen"
                else:
                    info[GAUSSIAN_ROUTE] = "/genecp"

                out_str = ""
                # gaussian flips out if you specify basis info for an element that
                # isn't on the molecule, so make sure the basis set has an element
                for basis in self.basis:
                    if basis.elements and not basis.user_defined:
                        out_str += " ".join([ele for ele in basis.elements])
                        out_str += " 0\n"
                        basis_name = Basis.get_gaussian(basis.name)
                        warning = basis.sanity_check_basis(
                            basis_name, "gaussian"
                        )
                        if warning:
                            warnings.append(warning)
                        out_str += basis_name
                        out_str += "\n****\n"

                for basis in self.basis:
                    if basis.elements:
                        if basis.user_defined:
                            if os.path.exists(basis.user_defined):
                                with open(basis.user_defined, "r") as f:
                                    lines = f.readlines()

                                i = 0
                                while i < len(lines):
                                    test = lines[i].strip()
                                    if not test or test.startswith("!"):
                                        i += 1
                                        continue

                                    ele = test.split()[0]
                                    while i < len(lines):
                                        if ele in basis.elements:
                                            out_str += lines[i]

                                        if lines[i].startswith("****"):
                                            break

                                        i += 1

                                    i += 1

                            # if the file does not exists, just insert the path as an @ file
                            else:
                                out_str += "@%s\n" % basis.user_defined

                info[GAUSSIAN_GEN_BASIS] = out_str

        if self.ecp is not None:
            out_str = ""
            for basis in self.ecp:
                if basis.elements and not basis.user_defined:
                    out_str += " ".join([ele for ele in basis.elements])
                    out_str += " 0\n"
                    basis_name = Basis.get_gaussian(basis.name)
                    warning = basis.sanity_check_basis(basis_name, "gaussian")
                    if warning:
                        warnings.append(warning)
                    out_str += basis_name
                    out_str += "\n"

            for basis in self.ecp:
                if basis.elements:
                    if basis.user_defined:
                        if os.path.exists(basis.user_defined):
                            with open(basis.user_defined, "r") as f:
                                lines = f.readlines()

                                i = 0
                                while i < len(lines):
                                    test = lines[i].strip()
                                    if not test or test.startswith("!"):
                                        i += 1
                                        continue

                                    ele = test.split()[0]
                                    while i < len(lines):
                                        if ele in basis.elements:
                                            out_str += lines[i]

                                        if lines[i].startswith("****"):
                                            break

                                        i += 1

                                    i += 1

                        else:
                            out_str += "@%s\n" % basis.user_defined

            info[GAUSSIAN_GEN_ECP] = out_str

            if self.basis is None:
                info[GAUSSIAN_ROUTE] = " Pseudo=Read"

        return info, warnings

    def get_orca_basis_info(self):
        """return dict for get_orca_header"""
        # TODO: warn if basis should be f12
        info = {ORCA_BLOCKS: {"basis": []}, ORCA_ROUTE: []}
        warnings = []

        first_basis = []

        if self.basis is not None:
            for basis in self.basis:
                if basis.elements:
                    if basis.aux_type is None:
                        if basis.aux_type not in first_basis:
                            if not basis.user_defined:
                                basis_name = Basis.get_orca(basis.name)
                                warning = Basis.sanity_check_basis(
                                    basis_name, "orca"
                                )
                                if warning:
                                    warnings.append(warning)
                                out_str = basis_name
                                info[ORCA_ROUTE].append(out_str)
                                first_basis.append(basis.aux_type)

                            else:
                                out_str = 'GTOName "%s"' % basis.user_defined
                                info[ORCA_BLOCKS]["basis"].append(out_str)
                                first_basis.append(basis.aux_type)

                        else:
                            for ele in basis.elements:
                                out_str = "newGTO            %-2s " % ele

                                if not basis.user_defined:
                                    basis_name = Basis.get_orca(basis.name)
                                    warning = Basis.sanity_check_basis(
                                        basis_name, "orca"
                                    )
                                    if warning:
                                        warnings.append(warning)
                                    out_str += '"%s" end' % basis_name
                                else:
                                    out_str += '"%s" end' % basis.user_defined

                                info[ORCA_BLOCKS]["basis"].append(out_str)

                    elif basis.aux_type.upper() == "C":
                        if basis.aux_type not in first_basis:
                            if not basis.user_defined:
                                basis_name = Basis.get_orca(basis.name) + "/C"
                                warning = Basis.sanity_check_basis(
                                    basis_name, "orca"
                                )
                                if warning:
                                    warnings.append(warning)
                                out_str = "%s" % basis_name
                                info[ORCA_ROUTE].append(out_str)
                                first_basis.append(basis.aux_type)

                            else:
                                out_str = (
                                    'AuxCGTOName "%s"' % basis.user_defined
                                )
                                info[ORCA_BLOCKS]["basis"].append(out_str)
                                first_basis.append(basis.aux_type)

                        else:
                            for ele in basis.elements:
                                out_str = "newAuxCGTO        %-2s " % ele

                                if not basis.user_defined:
                                    basis_name = (
                                        Basis.get_orca(basis.name) + "/C"
                                    )
                                    warning = Basis.sanity_check_basis(
                                        basis_name, "orca"
                                    )
                                    if warning:
                                        warnings.append(warning)

                                    out_str += '"%s" end' % basis_name
                                else:
                                    out_str += '"%s" end' % basis.user_defined

                                info[ORCA_BLOCKS]["basis"].append(out_str)

                    elif basis.aux_type.upper() == "J":
                        if basis.aux_type not in first_basis:
                            if not basis.user_defined:
                                basis_name = Basis.get_orca(basis.name) + "/J"
                                warning = Basis.sanity_check_basis(
                                    basis_name, "orca"
                                )
                                if warning:
                                    warnings.append(warning)
                                out_str = "%s" % basis_name
                                info[ORCA_ROUTE].append(out_str)
                                first_basis.append(basis.aux_type)

                            else:
                                out_str = (
                                    'AuxJGTOName "%s"' % basis.user_defined
                                )
                                info[ORCA_BLOCKS]["basis"].append(out_str)
                                first_basis.append(basis.aux_type)

                        else:
                            for ele in basis.elements:
                                out_str = "newAuxJGTO        %-2s " % ele

                                if not basis.user_defined:
                                    basis_name = (
                                        Basis.get_orca(basis.name) + "/J"
                                    )
                                    warning = Basis.sanity_check_basis(
                                        basis_name, "orca"
                                    )
                                    if warning:
                                        warnings.append(warning)

                                    out_str += '"%s" end' % basis_name
                                else:
                                    out_str += '"%s" end' % basis.user_defined

                                info[ORCA_BLOCKS]["basis"].append(out_str)

                    elif basis.aux_type.upper() == "JK":
                        if basis.aux_type not in first_basis:
                            if not basis.user_defined:
                                basis_name = Basis.get_orca(basis.name) + "/JK"
                                warning = Basis.sanity_check_basis(
                                    basis_name, "orca"
                                )
                                if warning:
                                    warnings.append(warning)
                                out_str = "%s" % basis_name
                                info[ORCA_ROUTE].append(out_str)
                                first_basis.append(basis.aux_type)

                            else:
                                out_str = (
                                    'AuxJKGTOName "%s"' % basis.user_defined
                                )
                                info[ORCA_BLOCKS]["basis"].append(out_str)
                                first_basis.append(basis.aux_type)

                        else:
                            for ele in basis.elements:
                                out_str = "newAuxJKGTO       %-2s " % ele

                                if not basis.user_defined:
                                    basis_name = (
                                        Basis.get_orca(basis.name) + "/JK"
                                    )
                                    warning = Basis.sanity_check_basis(
                                        basis_name, "orca"
                                    )
                                    if warning:
                                        warnings.append(warning)

                                    out_str += '"%s" end' % basis_name
                                else:
                                    out_str += '"%s" end' % basis.user_defined

                                info[ORCA_BLOCKS]["basis"].append(out_str)

                    elif basis.aux_type.upper() == "CABS":
                        if basis.aux_type not in first_basis:
                            if not basis.user_defined:
                                basis_name = (
                                    Basis.get_orca(basis.name) + "-CABS"
                                )
                                warning = Basis.sanity_check_basis(
                                    basis_name, "orca"
                                )
                                if warning:
                                    warnings.append(warning)
                                out_str = "%s" % basis_name
                                info[ORCA_ROUTE].append(out_str)
                                first_basis.append(basis.aux_type)

                            else:
                                out_str = (
                                    'CABSGTOName "%s"' % basis.user_defined
                                )
                                info[ORCA_BLOCKS]["basis"].append(out_str)
                                first_basis.append(basis.aux_type)

                        else:
                            for ele in basis.elements:
                                out_str = "newCABSGTO        %-2s " % ele

                                if not basis.user_defined:
                                    basis_name = (
                                        Basis.get_orca(basis.name) + "-CABS"
                                    )
                                    warning = Basis.sanity_check_basis(
                                        basis_name, "orca"
                                    )
                                    if warning:
                                        warnings.append(warning)

                                    out_str += '"%s" end' % basis_name
                                else:
                                    out_str += '"%s" end' % basis.user_defined

                                info[ORCA_BLOCKS]["basis"].append(out_str)

                    elif basis.aux_type.upper() == "OPTRI CABS":
                        if basis.aux_type not in first_basis:
                            if not basis.user_defined:
                                basis_name = (
                                    Basis.get_orca(basis.name) + "-OptRI"
                                )
                                warning = Basis.sanity_check_basis(
                                    basis_name, "orca"
                                )
                                if warning:
                                    warnings.append(warning)
                                out_str = "%s" % basis_name
                                info[ORCA_ROUTE].append(out_str)
                                first_basis.append(basis.aux_type)

                            else:
                                out_str = (
                                    'CABSGTOName "%s"' % basis.user_defined
                                )
                                info[ORCA_BLOCKS]["basis"].append(out_str)
                                first_basis.append(basis.aux_type)

                        else:
                            for ele in basis.elements:
                                out_str = "newCABSGTO        %-2s " % ele

                                if not basis.user_defined:
                                    basis_name = (
                                        Basis.get_orca(basis.name) + "-OptRI"
                                    )
                                    warning = Basis.sanity_check_basis(
                                        basis_name, "orca"
                                    )
                                    if warning:
                                        warnings.append(warning)

                                    out_str += '"%s" end' % basis_name
                                else:
                                    out_str += '"%s" end' % basis.user_defined

                                info[ORCA_BLOCKS]["basis"].append(out_str)

        if self.ecp is not None:
            for basis in self.ecp:
                if basis.elements and not basis.user_defined:
                    for ele in basis.elements:
                        out_str = "newECP            %-2s " % ele
                        basis_name = Basis.get_orca(basis.name)
                        warning = basis.sanity_check_basis(basis_name, "orca")
                        if warning:
                            warnings.append(warning)
                        out_str += '"%s" end' % basis_name

                        info[ORCA_BLOCKS]["basis"].append(out_str)

                elif basis.elements and basis.user_defined:
                    # TODO: check if this works
                    out_str = 'GTOName "%s"' % basis.user_defined

                    info[ORCA_BLOCKS]["basis"].append(out_str)

        return info, warnings

    def get_psi4_basis_info(self, sapt=False):
        """
        sapt: bool, use df_basis_sapt instead of df_basis_scf for jk basis
        return dict for get_psi4_header
        """
        out_str = dict()
        warnings = []

        if self.basis is not None:
            for basis in self.basis:
                aux_type = basis.aux_type
                basis_name = basis.get_psi4(basis.name)
                # JK and RI will try to guess what basis set is being requested
                # specifying "DF X" as the aux type will give more control
                # but the user will have to request -ri or -jkfit explicitly 
                if isinstance(aux_type, str):
                    aux_type = aux_type.upper()
                else:
                    aux_type = "BASIS"
                if aux_type == "JK":
                    aux_type = "df_basis_scf"
                    basis_name += "-jkfit"
                elif aux_type == "DF SCF":
                    aux_type = "df_basis_scf"
                elif aux_type == "DF SAPT":
                    aux_type = "df_basis_sapt"
                elif aux_type == "DF GUESS":
                    aux_type = "df_basis_guess"
                elif aux_type == "DF SAD":
                    aux_type = "df_basis_sad"
                elif aux_type == "DF MP2":
                    aux_type = "df_basis_mp2"
                elif aux_type == "DF DCT":
                    aux_type = "df_basis_dct"
                elif aux_type == "DF MCSCF":
                    aux_type = "df_basis_mcscf"
                elif aux_type == "DF CC":
                    aux_type = "df_basis_cc"
                elif aux_type == "DF ELST":
                    aux_type = "df_basis_elst"
                elif aux_type == "RI":
                    aux_type = "df_basis_%s"
                    if sapt:
                        aux_type = "df_basis_sapt"
                    if basis_name.lower() in ["sto-3g", "3-21g"]:
                        basis_name += "-rifit"
                    else:
                        basis_name += "-ri"

                if not basis.user_defined:
                    warning = basis.sanity_check_basis(basis_name, "psi4")
                    if warning:
                        warnings.append(warning)

                if aux_type not in out_str:
                    out_str[aux_type] = "%s {\n" % aux_type.lower()
                    out_str[aux_type] += "    assign    %s\n" % basis_name

                else:
                    for ele in basis.elements:
                        out_str[aux_type] += "    assign %-2s %s\n" % (
                            ele, basis_name
                        )

            if any(basis.user_defined for basis in self.basis):
                for basis in self.basis:
                    if basis.user_defined:
                        aux_type = basis.aux_type
                        if not aux_type:
                            aux_type = "BASIS"
                        aux_type = aux_type.upper()
                        if os.path.exists(basis.user_defined):
                            if aux_type not in out_str:
                                out_str[aux_type] = "%s {\n" % aux_type.lower()

                            out_str[aux_type] += "\n[%s]\n" % basis.name
                            with open(basis.user_defined, "r") as f:
                                lines = [
                                    line.rstrip()
                                    for line in f.readlines()
                                    if line.strip()
                                    and not line.startswith("!")
                                ]
                                out_str[aux_type] += "\n".join(lines)
                                out_str[aux_type] += "\n\n"

        s = "}\n\n".join(out_str.values())
        s += "}"

        info = {PSI4_BEFORE_GEOM: [s]}

        return info, warnings

    def check_for_elements(self, geometry, count_ecps=False):
        """checks to make sure each element is in a basis set"""
        warning = ""
        # assume all elements aren't in a basis set, remove from the list if they have a basis
        # need to check each type of aux basis
        elements = list(set([str(atom.element) for atom in geometry.atoms]))
        if self.basis is not None:
            elements_without_basis = {None: elements.copy()}
            for basis in self.basis:
                if basis.aux_type not in elements_without_basis:
                    elements_without_basis[basis.aux_type] = [
                        str(e) for e in elements
                    ]

                for element in basis.elements:
                    if element in elements_without_basis[basis.aux_type]:
                        elements_without_basis[basis.aux_type].remove(element)

            if count_ecps and self.ecp:
                for basis in self.basis:
                    if basis.aux_type != None and basis.aux_type != "no":
                        continue
                    for ecp in self.ecp:
                        for element in ecp.elements:
                            print("removing", element)
                            if element in elements_without_basis[basis.aux_type]:
                                elements_without_basis[basis.aux_type].remove(element)

            if any(
                elements_without_basis[aux]
                for aux in elements_without_basis.keys()
            ):
                for aux in elements_without_basis.keys():
                    if elements_without_basis[aux]:
                        if aux is not None and aux != "no":
                            warning += "%s ha%s no auxiliary %s basis; " % (
                                ", ".join(elements_without_basis[aux]),
                                "s"
                                if len(elements_without_basis[aux]) == 1
                                else "ve",
                                aux,
                            )

                        else:
                            warning += "%s ha%s no basis; " % (
                                ", ".join(elements_without_basis[aux]),
                                "s"
                                if len(elements_without_basis[aux]) == 1
                                else "ve",
                            )

                return warning.strip("; ")

            return None

    def get_qchem_basis_info(self, geom):
        """returns dict used by get_qchem_header with basis info"""
        info = {QCHEM_REM: dict(), QCHEM_SETTINGS: dict()}
        warnings = []
        
        if self.basis:
            no_aux_basis = [basis for basis in self.basis if not basis.aux_type]
            other_basis = [basis for basis in self.basis if basis not in no_aux_basis]
            aux_j_basis = [basis for basis in other_basis if basis.aux_type.lower() == "j"]
            aux_k_basis = [basis for basis in other_basis if basis.aux_type.lower() == "k"]
            aux_corr_basis = [basis for basis in other_basis if basis.aux_type.lower() == "corr"]
            aux_ri_basis = [basis for basis in other_basis if basis.aux_type.lower() == "ri"]
        else:
            no_aux_basis = []
            aux_j_basis = []
            aux_k_basis = []
            aux_corr_basis = []
            aux_ri_basis = []

        for basis_list, label in zip(
            [no_aux_basis, aux_j_basis, aux_k_basis, aux_corr_basis, aux_ri_basis],
            ["BASIS", "AUX_BASIS_J", "AUX_BASIS_K", "AUX_BASIS_CORR", "AUX_BASIS"],
        ):
            if basis_list:
                # check if we need to use gen or mixed:
                #    -a basis set is user-defined (stored in an external file e.g. from the BSE)
                #    -multiple basis sets
                if (
                    all([basis == basis_list[0] for basis in basis_list])
                    and not basis_list[0].user_defined
                ):
                    basis_name = Basis.get_qchem(basis_list[0].name)
                    warning = basis_list[0].sanity_check_basis(
                        basis_name, "qchem"
                    )
                    if warning:
                        warnings.append(warning)
                    info[QCHEM_REM][label] = "%s" % basis_name
                elif not any(basis.user_defined for basis in basis_list):
                    info[QCHEM_REM][label] = "General"
                else:
                    info[QCHEM_REM][label] = "MIXED"
    
                if any(x == info[QCHEM_REM][label] for x in ["MIXED", "General"]):
                    out_str = ""
                    for basis in basis_list:
                        if basis.elements and not basis.user_defined:
                            if info[QCHEM_REM][label] == "General":
                                for ele in basis.elements:
    
                                    out_str += "%-2s 0\n    " % ele
                                    basis_name = Basis.get_qchem(basis.name)
                                    warning = basis.sanity_check_basis(
                                        basis_name, "qchem"
                                    )
                                    if warning:
                                        warnings.append(warning)
                                    out_str += basis_name
                                    out_str += "\n    ****\n    "
    
                            else:
                                atoms = geom.find(ele)
                                for atom in atoms:
                                    out_str += "%s %i\n    " % (atom.element, geom.atoms.index(atom) + 1)
                                    basis_name = Basis.get_qchem(basis.name)
                                    warning = basis.sanity_check_basis(
                                        basis_name, "qchem"
                                    )
                                    if warning:
                                        warnings.append(warning)
                                    out_str += basis_name
                                    out_str += "\n    ****\n    "
    
                    for basis in basis_list:
                        if basis.elements and basis.user_defined:
                            if os.path.exists(basis.user_defined):
                                with open(basis.user_defined, "r") as f:
                                    lines = f.readlines()
                            
                                for element in basis.elements:
                                    atoms = geom.find(element)
                                    for atom in atoms:
                                        i = 0
                                        while i < len(lines):
                                            test = lines[i].strip()
                                            if not test or test.startswith("!") or test.startswith("$"):
                                                i += 1
                                                continue
            
                                            ele = test.split()[0]
                                            if ele == atom.element:
                                                out_str += "%s %i\n" % (ele, geom.atoms.index(atom))
                                                i += 1
                                                while i < len(lines):
                                                    if ele == atom.element:
                                                        out_str += lines[i]
                
                                                    if lines[i].startswith("****"):
                                                        break
                
                                                    i += 1
            
                                            i += 1
            
                                    # if the file does not exists, just insert the path as an @ file
                                    else:
                                        warnings.append("file not found: %s" % basis.user_defined)
    
                    info[QCHEM_SETTINGS][label.lower()] = [out_str.strip()]

        if self.ecp is not None and any(ecp.elements for ecp in self.ecp):
            # check if we need to use gen:
            #    -a basis set is user-defined (stored in an external file e.g. from the BSE)
            if (
                all([ecp == self.ecp[0] for ecp in self.ecp])
                and not self.ecp[0].user_defined
                and not self.basis
            ):
                basis_name = ECP.get_qchem(self.ecp[0].name)
                warning = self.ecp[0].sanity_check_basis(
                    basis_name, "qchem"
                )
                if warning:
                    warnings.append(warning)
                if QCHEM_REM not in info:
                    info[QCHEM_REM] = {"ECP": "%s" % basis_name}
                else:
                    info[QCHEM_REM]["ECP"] = "%s" % basis_name

            elif not any(basis.user_defined for basis in self.basis):
                if QCHEM_REM not in info:
                    info[QCHEM_REM] = {"ECP": "General"}
                else:
                    info[QCHEM_REM]["ECP"] = "General"
            else:
                if QCHEM_REM not in info:
                    info[QCHEM_REM] = {"ECP": "MIXED"}
                else:
                    info[QCHEM_REM]["ECP"] = "MIXED"

            if any(x == info[QCHEM_REM]["ECP"] for x in ["MIXED", "General"]):
                out_str = ""
                for basis in self.ecp:
                    if basis.elements and not basis.user_defined:
                        if info[QCHEM_REM]["ECP"] == "General":
                            for ele in basis.elements:
                                out_str += "%-2s 0\n    " % ele
                                basis_name = ECP.get_qchem(basis.name)
                                warning = basis.sanity_check_basis(
                                    basis_name, "qchem"
                                )
                                if warning:
                                    warnings.append(warning)
                                out_str += basis_name
                                out_str += "\n    ****\n    "

                        else:
                            atoms = geom.find(element)
                            for atom in atoms:
                                out_str += "%s %i\n    " % (atom.element, geom.atoms.index(atom) + 1)
                                basis_name = ECP.get_qchem(basis.name)
                                warning = basis.sanity_check_basis(
                                    basis_name, "qchem"
                                )
                                if warning:
                                    warnings.append(warning)
                                out_str += basis_name
                                out_str += "\n    ****\n    "

                for ecp in self.ecp:
                    if ecp.elements:
                        if ecp.user_defined:
                            if os.path.exists(ecp.user_defined):
                                with open(ecp.user_defined, "r") as f:
                                    lines = f.readlines()
                        
                                for element in ecp.elements:
                                    atoms = geom.find(element)
                                    for atom in atoms:
                                        i = 0
                                        while i < len(lines):
                                            test = lines[i].strip()
                                            if not test or test.startswith("!") or test.startswith("$"):
                                                i += 1
                                                continue
        
                                            ele = test.split()[0]
                                            if ele == atom.element:
                                                out_str += "%s %i\n" % (ele, geom.atoms.index(atom))
                                                i += 1
                                                while i < len(lines):
                                                    if ele == atom.element:
                                                        out_str += lines[i]
            
                                                    if lines[i].startswith("****"):
                                                        break
            
                                                    i += 1
        
                                            i += 1
        
                            # if the file does not exists, just insert the path as an @ file
                            else:
                                warnings.append("file not found: %s" % ecp.user_defined)
                
                if QCHEM_SETTINGS not in info:
                    info[QCHEM_SETTINGS] = {"ecp": [out_str.strip()]}
                else:
                    info[QCHEM_SETTINGS]["ecp"] = [out_str.strip()]


        return info, warnings



import configparser
import inspect
import logging
import os
import re

from AaronTools.const import AARONLIB, AARONTOOLS

config = configparser.ConfigParser(interpolation=None, comment_prefixes=("#"))
for filename in [
    os.path.join(AARONTOOLS, "config.ini"),
    os.path.join(AARONLIB, "config.ini"),
]:
    try:
        config.read(filename)
    except FileNotFoundError:
        continue
    except configparser.MissingSectionHeaderError:
        # add global options to default section
        with open(filename) as f:
            contents = "[DEFAULT]\n" + f.read()
        config.read_string(contents)

if "log_level" in config["DEFAULT"]:
    LOGLEVEL = config["DEFAULT"]["log_level"].upper()
else:
    LOGLEVEL = "WARNING"
if "print_citations" in config["DEFAULT"]:
    PRINT_CITATIONS = config["DEFAULT"].getboolean("print_citations")
else:
    PRINT_CITATIONS = False
try:
    SAVE_CITATIONS = config["DEFAULT"].getboolean("save_citations")
except ValueError:
    SAVE_CITATIONS = config["DEFAULT"].get("save_citations")
if SAVE_CITATIONS is False:
    SAVE_CITATIONS = None

logging.logThreads = 0
logging.logProcesses = 0

logging.captureWarnings(True)


class CustomFilter(logging.Filter):
    def __init__(self, name="", level=None, override=None, cite=False):
        super().__init__(name=name)

        self.level = logging.WARNING
        if isinstance(level, str):
            level = getattr(logging, level.upper())
        if level is not None:
            self.level = level
        self.override = {}
        if override is not None:
            self.override = override
        self.cite = cite

    def filter(self, record):
        if record.funcName == "citation":
            found = False
            for frame in reversed(inspect.stack()):
                if found:
                    record.funcName = frame.function
                    break
                if frame.function == "_callTestMethod":
                    found = True
            else:
                record.funcName = inspect.stack()[-2].function
            record.levelname = "CITATION"
            if not self.cite:
                return False
            self.parse_message(record)
            return True
        for level, func_list in self.override.items():
            if isinstance(level, str):
                level = getattr(logging, level.upper())
            if record.funcName not in func_list:
                continue
            if record.levelno < level:
                return False
            self.parse_message(record)
            return True
        if record.levelno < self.level:
            return False
        self.parse_message(record)
        return True

    def parse_message(self, record):
        """
        Formats message to print prettily to console
        """
        if isinstance(record.msg, str):
            record.msg = re.sub(
                "\n(\S)", lambda x: "\n  %s" % x.group(1), record.msg
            )
        msg = ["\n  "]
        for word in re.findall("\S+\s*", record.getMessage()):
            if len("".join(msg).split("\n")[-1]) + len(word) < 80:
                msg.append(word)
            else:
                msg.append("\n  {}".format(word))
        record.getMessage = lambda: "".join(msg)


class CitationHandler(logging.FileHandler):
    def __init__(self, filename, **kwargs):
        filename = os.path.expandvars(filename)
        if not os.path.exists(os.path.dirname(filename)):
            # might be trying to put citations in $AARONLIB, but user
            # didn't bother to set the environment variable and just
            # uses the default
            from AaronTools.const import AARONLIB

            if "$AARONLIB" in filename:
                filename = filename.replace("$AARONLIB", AARONLIB)
            elif "${AARONLIB}" in filename:
                filename = filename.replace("${AARONLIB}", AARONLIB)
            elif "%AARONLIB%" in filename:
                filename = filename.replace("%AARONLIB%", AARONLIB)
        super().__init__(filename, **kwargs)

    def emit(self, record):
        """
        Adds a record to the citation file if it's not already present
        """
        if record.levelname != "CITATION":
            return
        msg = record.msg.replace("\n  ", " ")
        record.getMessage = lambda: "".join(msg)
        # check for duplicates
        dupe = False
        with open(self.baseFilename) as f:
            for line in f.readlines():
                if line.strip() == self.format(record):
                    dupe = True
                    break
        if not dupe:
            super().emit(record)


class ATLogger(logging.Logger):
    def __init__(
        self, name, level=None, override=None, fmt=None, add_hdlrs=None
    ):
        """
        :level: the log level to use
        :override: dict(level=funcName) to override loglevel for certain funcitons
        :fmt: formatting string (optional)
        :add_hdlrs: list(str(handlerName)) or list(Handler())
        """
        super().__init__(name, level=1)
        if level is None:
            level = LOGLEVEL
        if isinstance(level, str):
            level = getattr(logging, level.upper())
        self.level = level
        if fmt is None:
            fmt = "%(levelname)s %(name)s.%(funcName)s %(message)s"

        formatter = logging.Formatter(fmt=fmt)
        handlers = [(logging.StreamHandler(), PRINT_CITATIONS)]
        if SAVE_CITATIONS is not None and os.access(SAVE_CITATIONS, os.W_OK):
            handlers += [(CitationHandler(SAVE_CITATIONS), True)]
        if add_hdlrs is not None:
            for hdlr in add_hdlrs:
                if isinstance(hdlr, str):
                    hdlr = getattr(logging, hdlr)
                    handlers.append((hdlr(), PRINT_CITATIONS))
                else:
                    handlers.append(hdlr, PRINT_CITATIONS)
        for hdlr, cite in handlers:
            hdlr.setFormatter(formatter)
            hdlr.addFilter(
                CustomFilter(
                    name=name, level=self.level, override=override, cite=cite
                )
            )
            self.addHandler(hdlr)

    def citation(self, msg, *args, **kwargs):
        self.info(msg, *args, **kwargs)


def getlogger(name=None, level=None, override=None, fmt=None):
    """
    Get the logger without using the class decorator
    :level: the log level to apply, defaults to WARNING
    :override: a dictionary of the form {new_level: function_name_list} will apply the
        `new_level` to log records produced from functions with names in
        `function_name_list`, eg:
            override={"DEBUG": ["some_function"]}
        will set the log level to DEBUG for any messages produced during the run of
        some_function()
    """
    if name is None:
        package = None
        for frame in reversed(inspect.stack()):
            res = inspect.getargvalues(frame.frame)
            if "__name__" in res.locals and name is None:
                name = res.locals["__name__"]
            if "__package__" in res.locals and package is None:
                package = res.locals["__package__"]
            if name is not None and package is not None:
                break
        name = "{}{}{}".format(
            name if name is not None else "",
            "." if package is not None else "",
            package if package is not None else "",
        )
    log = ATLogger(name, level=level, override=override, fmt=fmt)
    return log


def addlogger(cls):
    """
    Import this function and use it as a class decorator.
    Log messages using the created LOG class attribute.

    Useful class attributes to set that will be picked up by this decorator:
    :LOG: Will be set to the logger instance during class initialization
    :LOGLEVEL: Set this to use a different log level than what is in your config. Only
        do this for testing purposes, and do not include it when pushing commits to the
        master AaronTools branch.
    :LOGLEVEL_OVERRIDE: Use this dict to override the log level set in the config file
        for records originating in particular functions. Keys are log levels, values
        are lists of strings corresponding to function names (default: {})

    Example:
    ```
    from AaronTools import addlogger

    @addlogger
    class Someclass:
        LOG = None
        LOGLEVEL = "WARNING"
        LOGLEVEL_OVERRIDE = {"DEBUG": ["some_function"]}

        # this won't be printed b/c "INFO" < LOGLEVEL
        LOG.info("loading class")

        def some_function(self):
            # this message will be printed thanks to LOGLEVEL_OVERRIDE
            self.LOG.debug("function called")
    ```
    """
    name = "{}.{}".format(cls.__module__, cls.__name__)
    level = None
    if hasattr(cls, "LOGLEVEL") and cls.LOGLEVEL is not None:
        level = cls.LOGLEVEL
    override = None
    if hasattr(cls, "LOGLEVEL_OVERRIDE") and cls.LOGLEVEL_OVERRIDE is not None:
        override = cls.LOGLEVEL_OVERRIDE

    cls.LOG = ATLogger(name, level=level, override=override)
    return cls
